# Copyright 2018 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Requires Python 2.4+ and Openssl 1.0+
#

from __future__ import print_function

import contextlib
import os
import random
import re
import subprocess
import tempfile
import time
import threading

from nose.plugins.attrib import attr

from azurelinuxagent.common import conf
from azurelinuxagent.common.cgroup import AGENT_NAME_TELEMETRY, MetricsCounter, MetricValue, MetricsCategory, CpuCgroup
from azurelinuxagent.common.cgroupconfigurator import CGroupConfigurator, DisableCgroups
from azurelinuxagent.common.cgroupstelemetry import CGroupsTelemetry
from azurelinuxagent.common.event import WALAEventOperation
from azurelinuxagent.common.exception import CGroupsException, ExtensionError, ExtensionErrorCodes, \
    AgentMemoryExceededException
from azurelinuxagent.common.future import ustr
from azurelinuxagent.common.utils import shellutil, fileutil
from tests.common.mock_environment import MockCommand
from tests.common.mock_cgroup_environment import mock_cgroup_environment, UnitFilePaths
from tests.tools import AgentTestCase, patch, mock_sleep, i_am_root, data_dir, is_python_version_26_or_34, skip_if_predicate_true
from tests.utils.miscellaneous_tools import format_processes, wait_for


class CGroupConfiguratorSystemdTestCase(AgentTestCase):
    @classmethod
    def tearDownClass(cls):
        CGroupConfigurator._instance = None
        AgentTestCase.tearDownClass()

    @contextlib.contextmanager
    def _get_cgroup_configurator(self, initialize=True, enable=True, mock_commands=None):
        CGroupConfigurator._instance = None
        configurator = CGroupConfigurator.get_instance()
        CGroupsTelemetry.reset()
        with mock_cgroup_environment(self.tmp_dir) as mock_environment:
            if mock_commands is not None:
                for command in mock_commands:
                    mock_environment.add_command(command)
            configurator.mocks = mock_environment
            if initialize:
                if not enable:
                    with patch.object(configurator, "enable"):
                        configurator.initialize()
                else:
                    configurator.initialize()
            yield configurator

    def test_initialize_should_enable_cgroups(self):
        with self._get_cgroup_configurator() as configurator:
            self.assertTrue(configurator.enabled(), "cgroups were not enabled")

    def test_initialize_should_start_tracking_the_agent_cgroups(self):
        with self._get_cgroup_configurator() as configurator:
            tracked = CGroupsTelemetry._tracked

            self.assertTrue(configurator.enabled(), "Cgroups should be enabled")
            self.assertTrue(any(cg for cg in tracked.values() if cg.name == AGENT_NAME_TELEMETRY and 'cpu' in cg.path),
                "The Agent's CPU is not being tracked. Tracked: {0}".format(tracked))
            self.assertTrue(any(cg for cg in tracked.values() if cg.name == AGENT_NAME_TELEMETRY and 'memory' in cg.path),
                "The Agent's Memory is not being tracked. Tracked: {0}".format(tracked))

    def test_initialize_should_start_tracking_other_controllers_when_one_is_not_present(self):
        command_mocks = [MockCommand(r"^mount -t cgroup$",
'''cgroup on /sys/fs/cgroup/systemd type cgroup (rw,nosuid,nodev,noexec,relatime,xattr,name=systemd)
cgroup on /sys/fs/cgroup/rdma type cgroup (rw,nosuid,nodev,noexec,relatime,rdma)
cgroup on /sys/fs/cgroup/cpuset type cgroup (rw,nosuid,nodev,noexec,relatime,cpuset)
cgroup on /sys/fs/cgroup/net_cls,net_prio type cgroup (rw,nosuid,nodev,noexec,relatime,net_cls,net_prio)
cgroup on /sys/fs/cgroup/perf_event type cgroup (rw,nosuid,nodev,noexec,relatime,perf_event)
cgroup on /sys/fs/cgroup/hugetlb type cgroup (rw,nosuid,nodev,noexec,relatime,hugetlb)
cgroup on /sys/fs/cgroup/freezer type cgroup (rw,nosuid,nodev,noexec,relatime,freezer)
cgroup on /sys/fs/cgroup/pids type cgroup (rw,nosuid,nodev,noexec,relatime,pids)
cgroup on /sys/fs/cgroup/devices type cgroup (rw,nosuid,nodev,noexec,relatime,devices)
cgroup on /sys/fs/cgroup/cpu,cpuacct type cgroup (rw,nosuid,nodev,noexec,relatime,cpu,cpuacct)
cgroup on /sys/fs/cgroup/blkio type cgroup (rw,nosuid,nodev,noexec,relatime,blkio)
''')]
        with self._get_cgroup_configurator(mock_commands=command_mocks) as configurator:
            tracked = CGroupsTelemetry._tracked

            self.assertTrue(configurator.enabled(), "Cgroups should be enabled")
            self.assertFalse(any(cg for cg in tracked.values() if cg.name == 'walinuxagent.service' and 'memory' in cg.path),
                "The Agent's memory should not be tracked. Tracked: {0}".format(tracked))

    def test_initialize_should_not_enable_cgroups_when_the_cpu_and_memory_controllers_are_not_present(self):
        command_mocks = [MockCommand(r"^mount -t cgroup$",
'''cgroup on /sys/fs/cgroup/systemd type cgroup (rw,nosuid,nodev,noexec,relatime,xattr,name=systemd)
cgroup on /sys/fs/cgroup/rdma type cgroup (rw,nosuid,nodev,noexec,relatime,rdma)
cgroup on /sys/fs/cgroup/cpuset type cgroup (rw,nosuid,nodev,noexec,relatime,cpuset)
cgroup on /sys/fs/cgroup/net_cls,net_prio type cgroup (rw,nosuid,nodev,noexec,relatime,net_cls,net_prio)
cgroup on /sys/fs/cgroup/perf_event type cgroup (rw,nosuid,nodev,noexec,relatime,perf_event)
cgroup on /sys/fs/cgroup/hugetlb type cgroup (rw,nosuid,nodev,noexec,relatime,hugetlb)
cgroup on /sys/fs/cgroup/freezer type cgroup (rw,nosuid,nodev,noexec,relatime,freezer)
cgroup on /sys/fs/cgroup/pids type cgroup (rw,nosuid,nodev,noexec,relatime,pids)
cgroup on /sys/fs/cgroup/devices type cgroup (rw,nosuid,nodev,noexec,relatime,devices)
cgroup on /sys/fs/cgroup/blkio type cgroup (rw,nosuid,nodev,noexec,relatime,blkio)
''')]
        with self._get_cgroup_configurator(mock_commands=command_mocks) as configurator:
            tracked = CGroupsTelemetry._tracked

            self.assertFalse(configurator.enabled(), "Cgroups should not be enabled")
            self.assertEqual(len(tracked), 0, "No cgroups should be tracked. Tracked: {0}".format(tracked))

    def test_initialize_should_not_enable_cgroups_when_the_agent_is_not_in_the_system_slice(self):
        command_mocks = [MockCommand(r"^mount -t cgroup$",
'''cgroup on /sys/fs/cgroup/systemd type cgroup (rw,nosuid,nodev,noexec,relatime,xattr,name=systemd)
cgroup on /sys/fs/cgroup/rdma type cgroup (rw,nosuid,nodev,noexec,relatime,rdma)
cgroup on /sys/fs/cgroup/cpuset type cgroup (rw,nosuid,nodev,noexec,relatime,cpuset)
cgroup on /sys/fs/cgroup/net_cls,net_prio type cgroup (rw,nosuid,nodev,noexec,relatime,net_cls,net_prio)
cgroup on /sys/fs/cgroup/perf_event type cgroup (rw,nosuid,nodev,noexec,relatime,perf_event)
cgroup on /sys/fs/cgroup/hugetlb type cgroup (rw,nosuid,nodev,noexec,relatime,hugetlb)
cgroup on /sys/fs/cgroup/freezer type cgroup (rw,nosuid,nodev,noexec,relatime,freezer)
cgroup on /sys/fs/cgroup/pids type cgroup (rw,nosuid,nodev,noexec,relatime,pids)
cgroup on /sys/fs/cgroup/devices type cgroup (rw,nosuid,nodev,noexec,relatime,devices)
cgroup on /sys/fs/cgroup/blkio type cgroup (rw,nosuid,nodev,noexec,relatime,blkio)
''')]

        with self._get_cgroup_configurator(mock_commands=command_mocks) as configurator:
            tracked = CGroupsTelemetry._tracked
            agent_drop_in_file_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.cpu_quota)

            self.assertFalse(configurator.enabled(), "Cgroups should not be enabled")
            self.assertEqual(len(tracked), 0, "No cgroups should be tracked. Tracked: {0}".format(tracked))
            self.assertFalse(os.path.exists(agent_drop_in_file_cpu_quota), "{0} should not have been created".format(agent_drop_in_file_cpu_quota))

    def test_initialize_should_not_create_unit_files(self):
        with self._get_cgroup_configurator() as configurator:
            # get the paths to the mocked files
            azure_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.azure)
            extensions_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.vmextensions)
            agent_drop_in_file_slice = configurator.mocks.get_mapped_path(UnitFilePaths.slice)
            agent_drop_in_file_cpu_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.cpu_accounting)
            agent_drop_in_file_memory_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.memory_accounting)

            # The mock creates the slice unit files; delete them
            os.remove(azure_slice_unit_file)
            os.remove(extensions_slice_unit_file)

            # The service file for the agent includes settings for the slice and cpu accounting, but not for cpu quota; initialize()
            # should not create drop in files for the first 2, but it should create one the cpu quota
            self.assertFalse(os.path.exists(azure_slice_unit_file), "{0} should not have been created".format(azure_slice_unit_file))
            self.assertFalse(os.path.exists(extensions_slice_unit_file), "{0} should not have been created".format(extensions_slice_unit_file))
            self.assertFalse(os.path.exists(agent_drop_in_file_slice), "{0} should not have been created".format(agent_drop_in_file_slice))
            self.assertFalse(os.path.exists(agent_drop_in_file_cpu_accounting), "{0} should not have been created".format(agent_drop_in_file_cpu_accounting))
            self.assertFalse(os.path.exists(agent_drop_in_file_memory_accounting), "{0} should not have been created".format(agent_drop_in_file_memory_accounting))

    def test_initialize_should_create_unit_files_when_the_agent_service_file_is_not_updated(self):
        with self._get_cgroup_configurator(initialize=False) as configurator:
            # get the paths to the mocked files
            azure_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.azure)
            extensions_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.vmextensions)
            agent_drop_in_file_slice = configurator.mocks.get_mapped_path(UnitFilePaths.slice)
            agent_drop_in_file_cpu_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.cpu_accounting)
            agent_drop_in_file_memory_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.memory_accounting)

            # The mock creates the service and slice unit files; replace the former and delete the latter
            configurator.mocks.add_data_file(os.path.join(data_dir, 'init', "walinuxagent.service.previous"), UnitFilePaths.walinuxagent)
            os.remove(azure_slice_unit_file)
            os.remove(extensions_slice_unit_file)

            configurator.initialize()

            # The older service file for the agent did not include settings for the slice and cpu parameters; in that case, initialize() should
            # create drop in files to set those properties
            self.assertTrue(os.path.exists(azure_slice_unit_file), "{0} was not created".format(azure_slice_unit_file))
            self.assertTrue(os.path.exists(extensions_slice_unit_file), "{0} was not created".format(extensions_slice_unit_file))
            self.assertTrue(os.path.exists(agent_drop_in_file_slice), "{0} was not created".format(agent_drop_in_file_slice))
            self.assertTrue(os.path.exists(agent_drop_in_file_cpu_accounting), "{0} was not created".format(agent_drop_in_file_cpu_accounting))
            self.assertTrue(os.path.exists(agent_drop_in_file_memory_accounting), "{0} was not created".format(agent_drop_in_file_memory_accounting))

    def test_initialize_should_update_logcollector_memorylimit(self):
        with self._get_cgroup_configurator(initialize=False) as configurator:
            log_collector_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.logcollector)
            original_memory_limit = "MemoryLimit=30M"

            # The mock creates the slice unit file with memory limit
            configurator.mocks.add_data_file(os.path.join(data_dir, 'init', "azure-walinuxagent-logcollector.slice"),
                                             UnitFilePaths.logcollector)
            if not os.path.exists(log_collector_unit_file):
                raise Exception("{0} should have been created during test setup".format(log_collector_unit_file))
            if not fileutil.findre_in_file(log_collector_unit_file, original_memory_limit):
                raise Exception("MemoryLimit was not set correctly. Expected: {0}. Got:\n{1}".format(
                    original_memory_limit, fileutil.read_file(log_collector_unit_file)))

            configurator.initialize()

            # initialize() should update the unit file to remove the memory limit
            self.assertFalse(fileutil.findre_in_file(log_collector_unit_file, original_memory_limit),
                             "Log collector slice unit file was not updated correctly. Expected no memory limit. Got:\n{0}".format(
                                 fileutil.read_file(log_collector_unit_file)))

    def test_setup_extension_slice_should_create_unit_files(self):
        with self._get_cgroup_configurator() as configurator:
            # get the paths to the mocked files
            extension_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.extensionslice)

            configurator.setup_extension_slice(extension_name="Microsoft.CPlat.Extension", cpu_quota=5)

            expected_cpu_accounting = "CPUAccounting=yes"
            expected_cpu_quota_percentage = "5%"
            expected_memory_accounting = "MemoryAccounting=yes"

            self.assertTrue(os.path.exists(extension_slice_unit_file), "{0} was not created".format(extension_slice_unit_file))
            self.assertTrue(fileutil.findre_in_file(extension_slice_unit_file, expected_cpu_accounting),
                "CPUAccounting was not set correctly. Expected: {0}. Got:\n{1}".format(expected_cpu_accounting, fileutil.read_file(
                    extension_slice_unit_file)))
            self.assertTrue(fileutil.findre_in_file(extension_slice_unit_file, expected_cpu_quota_percentage),
                "CPUQuota was not set correctly. Expected: {0}. Got:\n{1}".format(expected_cpu_quota_percentage, fileutil.read_file(
                    extension_slice_unit_file)))
            self.assertTrue(fileutil.findre_in_file(extension_slice_unit_file, expected_memory_accounting),
                "MemoryAccounting was not set correctly. Expected: {0}. Got:\n{1}".format(expected_memory_accounting, fileutil.read_file(
                    extension_slice_unit_file)))

    def test_remove_extension_slice_should_remove_unit_files(self):
        with self._get_cgroup_configurator() as configurator:
            with patch("os.path.exists") as mock_path:
                mock_path.return_value = True
                # get the paths to the mocked files
                extension_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.extensionslice)

                CGroupsTelemetry._tracked['/sys/fs/cgroup/cpu,cpuacct/azure.slice/azure-vmextensions.slice/' \
                                          'azure-vmextensions-Microsoft.CPlat.Extension.slice'] = \
                    CpuCgroup('Microsoft.CPlat.Extension',
                              '/sys/fs/cgroup/cpu,cpuacct/azure.slice/azure-vmextensions.slice/azure-vmextensions-Microsoft.CPlat.Extension.slice')

                configurator.remove_extension_slice(extension_name="Microsoft.CPlat.Extension")

                tracked = CGroupsTelemetry._tracked

                self.assertFalse(
                    any(cg for cg in tracked.values() if cg.name == 'Microsoft.CPlat.Extension' and 'cpu' in cg.path),
                    "The extension's CPU is being tracked")
                self.assertFalse(
                    any(cg for cg in tracked.values() if cg.name == 'Microsoft.CPlat.Extension' and 'memory' in cg.path),
                    "The extension's Memory is being tracked")

            self.assertFalse(os.path.exists(extension_slice_unit_file), "{0} should not be present".format(extension_slice_unit_file))

    def test_enable_should_raise_cgroups_exception_when_cgroups_are_not_supported(self):
        with self._get_cgroup_configurator(enable=False) as configurator:
            with patch.object(configurator, "supported", return_value=False):
                with self.assertRaises(CGroupsException) as context_manager:
                    configurator.enable()
                self.assertIn("Attempted to enable cgroups, but they are not supported on the current platform", str(context_manager.exception))

    def test_enable_should_set_agent_cpu_quota_and_track_throttled_time(self):
        with self._get_cgroup_configurator(initialize=False) as configurator:
            agent_drop_in_file_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.cpu_quota)
            if os.path.exists(agent_drop_in_file_cpu_quota):
                raise Exception("{0} should not have been created during test setup".format(agent_drop_in_file_cpu_quota))

            configurator.initialize()

            expected_quota = "CPUQuota={0}%".format(conf.get_agent_cpu_quota())
            self.assertTrue(os.path.exists(agent_drop_in_file_cpu_quota), "{0} was not created".format(agent_drop_in_file_cpu_quota))
            self.assertTrue(
                fileutil.findre_in_file(agent_drop_in_file_cpu_quota, expected_quota),
                "CPUQuota was not set correctly. Expected: {0}. Got:\n{1}".format(expected_quota, fileutil.read_file(agent_drop_in_file_cpu_quota)))
            self.assertTrue(CGroupsTelemetry.get_track_throttled_time(), "Throttle time should be tracked")

    def test_enable_should_not_track_throttled_time_when_setting_the_cpu_quota_fails(self):
        with self._get_cgroup_configurator(initialize=False) as configurator:
            if CGroupsTelemetry.get_track_throttled_time():
                raise Exception("Test setup should not start tracking Throttle Time")

            configurator.mocks.add_file(UnitFilePaths.cpu_quota, Exception("A TEST EXCEPTION"))

            configurator.initialize()

            self.assertFalse(CGroupsTelemetry.get_track_throttled_time(), "Throttle time should not be tracked")

    def test_disable_should_reset_cpu_quota(self):
        with self._get_cgroup_configurator() as configurator:
            if len(CGroupsTelemetry._tracked) == 0:
                raise Exception("Test setup should have started tracking at least 1 cgroup (the agent's)")
            if not CGroupsTelemetry._track_throttled_time:
                raise Exception("Test setup should have started tracking Throttle Time")

            configurator.disable("UNIT TEST", DisableCgroups.AGENT)

            agent_drop_in_file_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.cpu_quota)
            self.assertTrue(os.path.exists(agent_drop_in_file_cpu_quota), "{0} was not created".format(agent_drop_in_file_cpu_quota))
            self.assertTrue(
                fileutil.findre_in_file(agent_drop_in_file_cpu_quota, "^CPUQuota=$"),
                "CPUQuota was not set correctly. Expected an empty value. Got:\n{0}".format(fileutil.read_file(agent_drop_in_file_cpu_quota)))
            self.assertEqual(len(CGroupsTelemetry._tracked), 1, "Memory cgroups should be tracked after disable. Tracking: {0}".format(CGroupsTelemetry._tracked))
            self.assertFalse(any(cg for cg in CGroupsTelemetry._tracked.values() if cg.name == 'walinuxagent.service' and 'cpu' in cg.path),
                "The Agent's cpu should not be tracked. Tracked: {0}".format(CGroupsTelemetry._tracked))

    def test_disable_should_reset_cpu_quota_for_all_cgroups(self):
        service_list = [
            {
                "name": "extension.service",
                "cpuQuotaPercentage": 5
            }
        ]
        extension_name = "Microsoft.CPlat.Extension"
        extension_services = {extension_name: service_list}
        with self._get_cgroup_configurator() as configurator:
            with patch.object(configurator, "get_extension_services_list", return_value=extension_services):
                # get the paths to the mocked files
                agent_drop_in_file_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.cpu_quota)
                extension_slice_unit_file = configurator.mocks.get_mapped_path(UnitFilePaths.extensionslice)
                extension_service_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_cpu_quota)

                configurator.setup_extension_slice(extension_name=extension_name, cpu_quota=5)
                configurator.set_extension_services_cpu_memory_quota(service_list)
                CGroupsTelemetry._tracked['/sys/fs/cgroup/cpu,cpuacct/system.slice/extension.service'] = \
                    CpuCgroup('extension.service', '/sys/fs/cgroup/cpu,cpuacct/system.slice/extension.service')
                CGroupsTelemetry._tracked['/sys/fs/cgroup/cpu,cpuacct/azure.slice/azure-vmextensions.slice/' \
                                          'azure-vmextensions-Microsoft.CPlat.Extension.slice'] = \
                    CpuCgroup('Microsoft.CPlat.Extension',
                              '/sys/fs/cgroup/cpu,cpuacct/azure.slice/azure-vmextensions.slice/azure-vmextensions-Microsoft.CPlat.Extension.slice')

                configurator.disable("UNIT TEST", DisableCgroups.ALL)

                self.assertTrue(os.path.exists(agent_drop_in_file_cpu_quota),
                                "{0} was not created".format(agent_drop_in_file_cpu_quota))
                self.assertTrue(
                    fileutil.findre_in_file(agent_drop_in_file_cpu_quota, "^CPUQuota=$"),
                    "CPUQuota was not set correctly. Expected an empty value. Got:\n{0}".format(
                        fileutil.read_file(agent_drop_in_file_cpu_quota)))
                self.assertTrue(os.path.exists(extension_slice_unit_file),
                                "{0} was not created".format(extension_slice_unit_file))
                self.assertTrue(
                    fileutil.findre_in_file(extension_slice_unit_file, "^CPUQuota=$"),
                    "CPUQuota was not set correctly. Expected an empty value. Got:\n{0}".format(
                        fileutil.read_file(extension_slice_unit_file)))
                self.assertTrue(os.path.exists(extension_service_cpu_quota),
                                "{0} was not created".format(extension_service_cpu_quota))
                self.assertTrue(
                    fileutil.findre_in_file(extension_service_cpu_quota, "^CPUQuota=$"),
                    "CPUQuota was not set correctly. Expected an empty value. Got:\n{0}".format(
                        fileutil.read_file(extension_service_cpu_quota)))
                self.assertEqual(len(CGroupsTelemetry._tracked), 0,
                                 "No cgroups should be tracked after disable. Tracking: {0}".format(
                                     CGroupsTelemetry._tracked))

    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_not_use_systemd_when_cgroups_are_not_enabled(self, _):
        with self._get_cgroup_configurator() as configurator:
            configurator.disable("UNIT TEST", DisableCgroups.ALL)

            with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", wraps=subprocess.Popen) as patcher:
                configurator.start_extension_command(
                    extension_name="Microsoft.Compute.TestExtension-1.2.3",
                    command="date",
                    cmd_name="test",
                    timeout=300,
                    shell=False,
                    cwd=self.tmp_dir,
                    env={},
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)

                command_calls = [args[0] for args, _ in patcher.call_args_list if len(args) > 0 and "date" in args[0]]
                self.assertEqual(len(command_calls), 1, "The test command should have been called exactly once [{0}]".format(command_calls))
                self.assertNotIn("systemd-run", command_calls[0], "The command should not have been invoked using systemd")
                self.assertEqual(command_calls[0], "date", "The command line should not have been modified")

    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_use_systemd_run_when_cgroups_are_enabled(self, _):
        with self._get_cgroup_configurator() as configurator:
            with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", wraps=subprocess.Popen) as popen_patch:
                configurator.start_extension_command(
                    extension_name="Microsoft.Compute.TestExtension-1.2.3",
                    command="the-test-extension-command",
                    cmd_name="test",
                    timeout=300,
                    shell=False,
                    cwd=self.tmp_dir,
                    env={},
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)

                command_calls = [args[0] for (args, _) in popen_patch.call_args_list if "the-test-extension-command" in args[0]]

                self.assertEqual(len(command_calls), 1, "The test command should have been called exactly once [{0}]".format(command_calls))
                self.assertIn("systemd-run", command_calls[0], "The extension should have been invoked using systemd")

    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_start_tracking_the_extension_cgroups(self, _):
        # CPU usage is initialized when we begin tracking a CPU cgroup; since this test does not retrieve the
        # CPU usage, there is no need for initialization
        with self._get_cgroup_configurator() as configurator:
            configurator.start_extension_command(
                extension_name="Microsoft.Compute.TestExtension-1.2.3",
                command="test command",
                cmd_name="test",
                timeout=300,
                shell=False,
                cwd=self.tmp_dir,
                env={},
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)

        tracked = CGroupsTelemetry._tracked

        self.assertTrue(
            any(cg for cg in tracked.values() if cg.name == 'Microsoft.Compute.TestExtension-1.2.3' and 'cpu' in cg.path),
            "The extension's CPU is not being tracked")
        self.assertTrue(
            any(cg for cg in tracked.values() if cg.name == 'Microsoft.Compute.TestExtension-1.2.3' and 'memory' in cg.path),
            "The extension's Memory is not being tracked")

    def test_start_extension_command_should_raise_an_exception_when_the_command_cannot_be_started(self):
        with self._get_cgroup_configurator() as configurator:
            original_popen = subprocess.Popen

            def mock_popen(command_arg, *args, **kwargs):
                if "test command" in command_arg:
                    raise Exception("A TEST EXCEPTION")
                return original_popen(command_arg, *args, **kwargs)

            with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", side_effect=mock_popen):
                with self.assertRaises(Exception) as context_manager:
                    configurator.start_extension_command(
                        extension_name="Microsoft.Compute.TestExtension-1.2.3",
                        command="test command",
                        cmd_name="test",
                        timeout=300,
                        shell=False,
                        cwd=self.tmp_dir,
                        env={},
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE)

                    self.assertIn("A TEST EXCEPTION", str(context_manager.exception))

    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_disable_cgroups_and_invoke_the_command_directly_if_systemd_fails(self, _):
        with self._get_cgroup_configurator() as configurator:

            configurator.mocks.add_command(MockCommand("systemd-run", return_value=1, stdout='', stderr='Failed to start transient scope unit: syntax error'))

            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as output_file:
                with patch("azurelinuxagent.common.cgroupconfigurator.add_event") as mock_add_event:
                    with patch("subprocess.Popen", wraps=subprocess.Popen) as popen_patch:
                        CGroupsTelemetry.reset()

                        command = "echo TEST_OUTPUT"

                        command_output = configurator.start_extension_command(
                            extension_name="Microsoft.Compute.TestExtension-1.2.3",
                            command=command,
                            cmd_name="test",
                            timeout=300,
                            shell=True,
                            cwd=self.tmp_dir,
                            env={},
                            stdout=output_file,
                            stderr=output_file)

                        self.assertFalse(configurator.enabled(), "Cgroups should have been disabled")

                        disabled_events = [kwargs for _, kwargs in mock_add_event.call_args_list if kwargs['op'] == WALAEventOperation.CGroupsDisabled]

                        self.assertTrue(len(disabled_events) == 1, "Exactly one CGroupsDisabled telemetry event should have been issued. Found: {0}".format(disabled_events))
                        self.assertIn("Failed to start Microsoft.Compute.TestExtension-1.2.3 using systemd-run",
                                      disabled_events[0]['message'],
                                      "The systemd-run failure was not included in the telemetry message")
                        self.assertEqual(False, disabled_events[0]['is_success'], "The telemetry event should indicate a failure")

                        extension_calls = [args[0] for (args, _) in popen_patch.call_args_list if command in args[0]]

                        self.assertEqual(2, len(extension_calls), "The extension should have been invoked exactly twice")
                        self.assertIn("systemd-run", extension_calls[0],
                                      "The first call to the extension should have used systemd")
                        self.assertEqual(command, extension_calls[1],
                                          "The second call to the extension should not have used systemd")

                        self.assertEqual(len(CGroupsTelemetry._tracked), 0, "No cgroups should have been created")

                        self.assertIn("TEST_OUTPUT\n", command_output, "The test output was not captured")

    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_disable_cgroups_and_invoke_the_command_directly_if_systemd_times_out(self, _):
        with self._get_cgroup_configurator() as configurator:
            # Systemd has its own internal timeout which is shorter than what we define for extension operation timeout.
            # When systemd times out, it will write a message to stderr and exit with exit code 1.
            # In that case, we will internally recognize the failure due to the non-zero exit code, not as a timeout.
            configurator.mocks.add_command(MockCommand("systemd-run", return_value=1, stdout='', stderr='Failed to start transient scope unit: Connection timed out'))

            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stdout:
                with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stderr:
                    with patch("subprocess.Popen", wraps=subprocess.Popen) as popen_patch:
                        CGroupsTelemetry.reset()

                        configurator.start_extension_command(
                            extension_name="Microsoft.Compute.TestExtension-1.2.3",
                            command="echo 'success'",
                            cmd_name="test",
                            timeout=300,
                            shell=True,
                            cwd=self.tmp_dir,
                            env={},
                            stdout=stdout,
                            stderr=stderr)

                        self.assertFalse(configurator.enabled(), "Cgroups should have been disabled")

                        extension_calls = [args[0] for (args, _) in popen_patch.call_args_list if "echo 'success'" in args[0]]
                        self.assertEqual(2, len(extension_calls), "The extension should have been called twice. Got: {0}".format(extension_calls))
                        self.assertIn("systemd-run", extension_calls[0], "The first call to the extension should have used systemd")
                        self.assertNotIn("systemd-run", extension_calls[1], "The second call to the extension should not have used systemd")

                        self.assertEqual(len(CGroupsTelemetry._tracked), 0, "No cgroups should have been created")

    @skip_if_predicate_true(is_python_version_26_or_34, "Disabled on Python 2.6 and 3.4 for now. Need to revisit to fix it")
    @attr('requires_sudo')
    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_not_use_fallback_option_if_extension_fails(self, *args):
        self.assertTrue(i_am_root(), "Test does not run when non-root")

        with self._get_cgroup_configurator() as configurator:
            pass  # release the mocks used to create the test CGroupConfigurator so that they do not conflict the mock Popen below

        command = "ls folder_does_not_exist"

        with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stdout:
            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stderr:
                with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", wraps=subprocess.Popen) as popen_patch:
                    with self.assertRaises(ExtensionError) as context_manager:
                        configurator.start_extension_command(
                            extension_name="Microsoft.Compute.TestExtension-1.2.3",
                            command=command,
                            cmd_name="test",
                            timeout=300,
                            shell=True,
                            cwd=self.tmp_dir,
                            env={},
                            stdout=stdout,
                            stderr=stderr)

                    extension_calls = [args[0] for (args, _) in popen_patch.call_args_list if command in args[0]]

                    self.assertEqual(1, len(extension_calls), "The extension should have been invoked exactly once")
                    self.assertIn("systemd-run", extension_calls[0],
                                  "The first call to the extension should have used systemd")

                    self.assertEqual(context_manager.exception.code, ExtensionErrorCodes.PluginUnknownFailure)
                    self.assertIn("Non-zero exit code", ustr(context_manager.exception))
                    # The scope name should appear in the process output since systemd-run was invoked and stderr
                    # wasn't truncated.
                    self.assertIn("Running scope as unit", ustr(context_manager.exception))

    @skip_if_predicate_true(is_python_version_26_or_34, "Disabled on Python 2.6 and 3.4 for now. Need to revisit to fix it")
    @attr('requires_sudo')
    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    @patch("azurelinuxagent.common.utils.extensionprocessutil.TELEMETRY_MESSAGE_MAX_LEN", 5)
    def test_start_extension_command_should_not_use_fallback_option_if_extension_fails_with_long_output(self, *args):
        self.assertTrue(i_am_root(), "Test does not run when non-root")

        with self._get_cgroup_configurator() as configurator:
            pass  # release the mocks used to create the test CGroupConfigurator so that they do not conflict the mock Popen below

        long_output = "a"*20  # large enough to ensure both stdout and stderr are truncated
        long_stdout_stderr_command = "echo {0} && echo {0} >&2 && ls folder_does_not_exist".format(long_output)

        with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stdout:
            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stderr:
                with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", wraps=subprocess.Popen) as popen_patch:
                    with self.assertRaises(ExtensionError) as context_manager:
                        configurator.start_extension_command(
                            extension_name="Microsoft.Compute.TestExtension-1.2.3",
                            command=long_stdout_stderr_command,
                            cmd_name="test",
                            timeout=300,
                            shell=True,
                            cwd=self.tmp_dir,
                            env={},
                            stdout=stdout,
                            stderr=stderr)

                    extension_calls = [args[0] for (args, _) in popen_patch.call_args_list if long_stdout_stderr_command in args[0]]

                    self.assertEqual(1, len(extension_calls), "The extension should have been invoked exactly once")
                    self.assertIn("systemd-run", extension_calls[0],
                                  "The first call to the extension should have used systemd")

                    self.assertEqual(context_manager.exception.code, ExtensionErrorCodes.PluginUnknownFailure)
                    self.assertIn("Non-zero exit code", ustr(context_manager.exception))
                    # stdout and stderr should have been truncated, so the scope name doesn't appear in stderr
                    # even though systemd-run ran
                    self.assertNotIn("Running scope as unit", ustr(context_manager.exception))

    @attr('requires_sudo')
    def test_start_extension_command_should_not_use_fallback_option_if_extension_times_out(self, *args):  # pylint: disable=unused-argument
        self.assertTrue(i_am_root(), "Test does not run when non-root")

        with self._get_cgroup_configurator() as configurator:
            pass  # release the mocks used to create the test CGroupConfigurator so that they do not conflict the mock Popen below

        with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stdout:
            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stderr:
                with patch("azurelinuxagent.common.utils.extensionprocessutil.wait_for_process_completion_or_timeout",
                           return_value=[True, None, 0]):
                    with patch("azurelinuxagent.common.cgroupapi.SystemdCgroupsApi._is_systemd_failure",
                               return_value=False):
                        with self.assertRaises(ExtensionError) as context_manager:
                            configurator.start_extension_command(
                                extension_name="Microsoft.Compute.TestExtension-1.2.3",
                                command="date",
                                cmd_name="test",
                                timeout=300,
                                shell=True,
                                cwd=self.tmp_dir,
                                env={},
                                stdout=stdout,
                                stderr=stderr)

                        self.assertEqual(context_manager.exception.code, ExtensionErrorCodes.PluginHandlerScriptTimedout)
                        self.assertIn("Timeout", ustr(context_manager.exception))

    @patch('time.sleep', side_effect=lambda _: mock_sleep())
    def test_start_extension_command_should_capture_only_the_last_subprocess_output(self, _):
        with self._get_cgroup_configurator() as configurator:
            pass  # release the mocks used to create the test CGroupConfigurator so that they do not conflict the mock Popen below

        original_popen = subprocess.Popen

        def mock_popen(command, *args, **kwargs):
            # Inject a syntax error to the call
            
            # Popen can accept both strings and lists, handle both here.
            if isinstance(command, str):
                systemd_command = command.replace('systemd-run', 'systemd-run syntax_error')
            elif isinstance(command, list) and command[0] == 'systemd-run':
                systemd_command = ['systemd-run', 'syntax_error'] + command[1:]

            return original_popen(systemd_command, *args, **kwargs)

        expected_output = "[stdout]\n{0}\n\n\n[stderr]\n"

        with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stdout:
            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stderr:
                with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", side_effect=mock_popen):
                    # We expect this call to fail because of the syntax error
                    process_output = configurator.start_extension_command(
                        extension_name="Microsoft.Compute.TestExtension-1.2.3",
                        command="echo 'very specific test message'",
                        cmd_name="test",
                        timeout=300,
                        shell=True,
                        cwd=self.tmp_dir,
                        env={},
                        stdout=stdout,
                        stderr=stderr)

                    self.assertEqual(expected_output.format("very specific test message"), process_output)

    def test_it_should_set_extension_services_cpu_memory_quota(self):
        service_list = [
            {
                "name": "extension.service",
                "cpuQuotaPercentage": 5
            }
        ]
        with self._get_cgroup_configurator() as configurator:
            # get the paths to the mocked files
            extension_service_cpu_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_cpu_accounting)
            extension_service_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_cpu_quota)

            configurator.set_extension_services_cpu_memory_quota(service_list)
            expected_cpu_accounting = "CPUAccounting=yes"
            expected_cpu_quota_percentage = "CPUQuota=5%"

            # create drop in files to set those properties
            self.assertTrue(os.path.exists(extension_service_cpu_accounting), "{0} was not created".format(extension_service_cpu_accounting))
            self.assertTrue(
                fileutil.findre_in_file(extension_service_cpu_accounting, expected_cpu_accounting),
                "CPUAccounting was not enabled. Expected: {0}. Got:\n{1}".format(expected_cpu_accounting, fileutil.read_file(extension_service_cpu_accounting)))

            self.assertTrue(os.path.exists(extension_service_cpu_quota), "{0} was not created".format(extension_service_cpu_quota))
            self.assertTrue(
                fileutil.findre_in_file(extension_service_cpu_quota, expected_cpu_quota_percentage),
                "CPUQuota was not set. Expected: {0}. Got:\n{1}".format(expected_cpu_quota_percentage, fileutil.read_file(extension_service_cpu_quota)))

    def test_it_should_set_extension_services_when_quotas_not_defined(self):
        service_list = [
            {
                "name": "extension.service"
            }
        ]
        with self._get_cgroup_configurator() as configurator:
            # get the paths to the mocked files
            extension_service_cpu_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_cpu_accounting)
            extension_service_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_cpu_quota)
            extension_service_memory_accounting = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_memory_accounting)
            extension_service_memory_quota = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_memory_limit)

            configurator.set_extension_services_cpu_memory_quota(service_list)

            self.assertTrue(os.path.exists(extension_service_cpu_accounting),
                            "{0} was not created".format(extension_service_cpu_accounting))
            self.assertFalse(os.path.exists(extension_service_cpu_quota),
                            "{0} should not have been created during setup".format(extension_service_cpu_quota))

            self.assertTrue(os.path.exists(extension_service_memory_accounting),
                            "{0} was not created".format(extension_service_memory_accounting))
            self.assertFalse(os.path.exists(extension_service_memory_quota),
                            "{0} should not have been created during setup".format(extension_service_memory_quota))

    def test_it_should_start_tracking_extension_services_cgroups(self):
        service_list = [
            {
                "name": "extension.service"
            }
        ]
        with self._get_cgroup_configurator() as configurator:
            configurator.start_tracking_extension_services_cgroups(service_list)

        tracked = CGroupsTelemetry._tracked

        self.assertTrue(
            any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'cpu' in cg.path),
            "The extension service's CPU is not being tracked")
        self.assertTrue(
            any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'memory' in cg.path),
            "The extension service's Memory is not being tracked")

    def test_it_should_stop_tracking_extension_services_cgroups(self):
        service_list = [
            {
                "name": "extension.service"
            }
        ]

        with self._get_cgroup_configurator() as configurator:
            with patch("os.path.exists") as mock_path:
                mock_path.return_value = True
                CGroupsTelemetry.track_cgroup(CpuCgroup('extension.service', '/sys/fs/cgroup/cpu,cpuacct/system.slice/extension.service'))
                configurator.stop_tracking_extension_services_cgroups(service_list)

                tracked = CGroupsTelemetry._tracked

                self.assertFalse(
                    any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'cpu' in cg.path),
                    "The extension service's CPU is being tracked")
                self.assertFalse(
                    any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'memory' in cg.path),
                    "The extension service's Memory is being tracked")

    def test_it_should_remove_extension_services_drop_in_files(self):
        service_list = [
            {
                "name": "extension.service",
                "cpuQuotaPercentage": 5
            }
        ]
        with self._get_cgroup_configurator() as configurator:
            extension_service_cpu_accounting = configurator.mocks.get_mapped_path(
            UnitFilePaths.extension_service_cpu_accounting)
            extension_service_cpu_quota = configurator.mocks.get_mapped_path(UnitFilePaths.extension_service_cpu_quota)
            extension_service_memory_accounting = configurator.mocks.get_mapped_path(
            UnitFilePaths.extension_service_memory_accounting)
            configurator.remove_extension_services_drop_in_files(service_list)
            self.assertFalse(os.path.exists(extension_service_cpu_accounting),
                            "{0} should not have been created".format(extension_service_cpu_accounting))
            self.assertFalse(os.path.exists(extension_service_cpu_quota),
                            "{0} should not have been created".format(extension_service_cpu_quota))
            self.assertFalse(os.path.exists(extension_service_memory_accounting),
                            "{0} should not have been created".format(extension_service_memory_accounting))

    def test_it_should_start_tracking_unit_cgroups(self):

        with self._get_cgroup_configurator() as configurator:
            configurator.start_tracking_unit_cgroups("extension.service")

        tracked = CGroupsTelemetry._tracked

        self.assertTrue(
            any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'cpu' in cg.path),
            "The extension service's CPU is not being tracked")

        self.assertTrue(
            any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'memory' in cg.path),
            "The extension service's Memory is not being tracked")

    def test_it_should_stop_tracking_unit_cgroups(self):

        def side_effect(path):
            if path == '/sys/fs/cgroup/cpu,cpuacct/system.slice/extension.service':
                return True
            return False

        with self._get_cgroup_configurator() as configurator:
            with patch("os.path.exists") as mock_path:
                mock_path.side_effect = side_effect
                CGroupsTelemetry._tracked['/sys/fs/cgroup/cpu,cpuacct/system.slice/extension.service'] = \
                    CpuCgroup('extension.service', '/sys/fs/cgroup/cpu,cpuacct/system.slice/extension.service')
                configurator.stop_tracking_unit_cgroups("extension.service")

                tracked = CGroupsTelemetry._tracked

                self.assertFalse(
                    any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'cpu' in cg.path),
                    "The extension service's CPU is being tracked")
                self.assertFalse(
                    any(cg for cg in tracked.values() if cg.name == 'extension.service' and 'memory' in cg.path),
                    "The extension service's Memory is being tracked")

    def test_check_processes_in_agent_cgroup_should_raise_a_cgroups_exception_when_there_are_unexpected_processes_in_the_agent_cgroup(self):
        with self._get_cgroup_configurator() as configurator:
            pass  # release the mocks used to create the test CGroupConfigurator so that they do not conflict the mock Popen below

        # The test script recursively creates a given number of descendant processes, then it blocks until the
        # 'stop_file' exists. It produces an output file containing the PID of each descendant process.
        test_script = os.path.join(self.tmp_dir, "create_processes.sh")
        stop_file = os.path.join(self.tmp_dir, "create_processes.stop")
        AgentTestCase.create_script(test_script, """
#!/usr/bin/env bash
set -euo pipefail

if [[ $# != 2 ]]; then
    echo "Usage: $0 <output_file> <count>"
    exit 1
fi

echo $$ >> $1

if [[ $2 > 1 ]]; then
    $0 $1 $(($2 - 1))
else
    timeout 30s /usr/bin/env bash -c "while ! [[ -f {0} ]]; do sleep 0.25s; done"
fi

exit 0
""".format(stop_file))

        number_of_descendants = 3

        def wait_for_processes(processes_file):
            def _all_present():
                if os.path.exists(processes_file):
                    with open(processes_file, "r") as file_stream:
                        _all_present.processes = [int(process) for process in file_stream.read().split()]
                return len(_all_present.processes) >= number_of_descendants
            _all_present.processes = []

            if not wait_for(_all_present):
                raise Exception("Timeout waiting for processes. Expected {0}; got: {1}".format(
                    number_of_descendants, format_processes(_all_present.processes)))

            return _all_present.processes

        threads = []

        try:
            #
            # Start the processes that will be used by the test. We use two sets of processes: the first set simulates a command executed by the agent
            # (e.g. iptables) and its child processes, if any. The second set of processes simulates an extension.
            #
            agent_command_output = os.path.join(self.tmp_dir, "agent_command.pids")
            agent_command = threading.Thread(target=lambda: shellutil.run_command([test_script, agent_command_output, str(number_of_descendants)]))
            agent_command.start()
            threads.append(agent_command)
            agent_command_processes = wait_for_processes(agent_command_output)

            extension_output = os.path.join(self.tmp_dir, "extension.pids")

            def start_extension():
                original_sleep = time.sleep
                original_popen = subprocess.Popen

                # Extensions are started using systemd-run; mock Popen to remove the call to systemd-run; the test script creates a couple of
                # child processes, which would simulate the extension's processes.
                def mock_popen(command, *args, **kwargs):
                    match = re.match(r"^systemd-run --property=CPUAccounting=no --property=MemoryAccounting=no --unit=[^\s]+ --scope --slice=[^\s]+ (.+)", command)
                    is_systemd_run = match is not None
                    if is_systemd_run:
                        command = match.group(1)
                    process = original_popen(command, *args, **kwargs)
                    if is_systemd_run:
                        start_extension.systemd_run_pid = process.pid
                    return process

                with patch('time.sleep', side_effect=lambda _: original_sleep(0.1)):  # start_extension_command has a small delay; skip it
                    with patch("azurelinuxagent.common.cgroupapi.subprocess.Popen", side_effect=mock_popen):
                        with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stdout:
                            with tempfile.TemporaryFile(dir=self.tmp_dir, mode="w+b") as stderr:
                                configurator.start_extension_command(
                                    extension_name="TestExtension",
                                    command="{0} {1} {2}".format(test_script, extension_output, number_of_descendants),
                                    cmd_name="test",
                                    timeout=30,
                                    shell=True,
                                    cwd=self.tmp_dir,
                                    env={},
                                    stdout=stdout,
                                    stderr=stderr)
            start_extension.systemd_run_pid = None

            extension = threading.Thread(target=start_extension)
            extension.start()
            threads.append(extension)
            extension_processes = wait_for_processes(extension_output)

            #
            # check_processes_in_agent_cgroup uses shellutil and the cgroups api to get the commands that are currently running;
            # wait for all the processes to show up
            #
            if not wait_for(lambda: len(shellutil.get_running_commands()) > 0 and len(configurator._cgroups_api.get_systemd_run_commands()) > 0):
                raise Exception("Timeout while attempting to track the child commands")

            #
            # Verify that check_processes_in_agent_cgroup raises when there are unexpected processes in the agent's cgroup.
            #
            # For the agent's processes, we use the current process and its parent (in the actual agent these would be the daemon and the extension
            # handler), and the commands started by the agent.
            #
            # For other processes, we use process 1, a process that already completed, and an extension. Note that extensions are started using
            # systemd-run and the process for that commands belongs to the agent's cgroup but the processes for the extension should be in a
            # different cgroup
            #
            def get_completed_process():
                random.seed()
                completed = random.randint(1000, 10000)
                while os.path.exists("/proc/{0}".format(completed)):  # ensure we do not use an existing process
                    completed = random.randint(1000, 10000)
                return completed

            agent_processes = [os.getppid(), os.getpid()] + agent_command_processes + [start_extension.systemd_run_pid]
            other_processes = [1, get_completed_process()] + extension_processes

            with patch("azurelinuxagent.common.cgroupconfigurator.CGroupsApi.get_processes_in_cgroup", return_value=agent_processes + other_processes):
                with self.assertRaises(CGroupsException) as context_manager:
                    configurator._check_processes_in_agent_cgroup()

                # The list of processes in the message is an array of strings: "['foo', ..., 'bar']"
                message = ustr(context_manager.exception)
                search = re.search(r'unexpected processes: \[(?P<processes>.+)\]', message)
                self.assertIsNotNone(search, "The event message is not in the expected format: {0}".format(message))
                reported = search.group('processes').split(',')

                self.assertEqual(
                    len(other_processes), len(reported),
                    "An incorrect number of processes was reported. Expected: {0} Got: {1}".format(format_processes(other_processes), reported))
                for pid in other_processes:
                    self.assertTrue(
                        any("[PID: {0}]".format(pid) in reported_process for reported_process in reported),
                        "Process {0} was not reported. Got: {1}".format(format_processes([pid]), reported))
        finally:
            # create the file that stops the test processes and wait for them to complete
            open(stop_file, "w").close()
            for thread in threads:
                thread.join(timeout=5)

    def test_check_agent_throttled_time_should_raise_a_cgroups_exception_when_the_threshold_is_exceeded(self):
        metrics = [MetricValue(MetricsCategory.CPU_CATEGORY, MetricsCounter.THROTTLED_TIME, AGENT_NAME_TELEMETRY, conf.get_agent_cpu_throttled_time_threshold() + 1)]

        with self.assertRaises(CGroupsException) as context_manager:
            CGroupConfigurator._Impl._check_agent_throttled_time(metrics)

        self.assertIn("The agent has been throttled", ustr(context_manager.exception), "An incorrect exception was raised")

    def test_check_cgroups_should_disable_cgroups_when_a_check_fails(self):
        with self._get_cgroup_configurator() as configurator:
            checks = ["_check_processes_in_agent_cgroup", "_check_agent_throttled_time"]
            for method_to_fail in checks:
                patchers = []
                try:
                    # mock 'method_to_fail' to raise an exception and the rest to do nothing
                    for method_to_mock in checks:
                        side_effect = CGroupsException(method_to_fail) if method_to_mock == method_to_fail else lambda *_: None
                        p = patch.object(configurator, method_to_mock, side_effect=side_effect)
                        patchers.append(p)
                        p.start()

                    with patch("azurelinuxagent.common.cgroupconfigurator.add_event") as add_event:
                        configurator.enable()

                        tracked_metrics = [
                            MetricValue(MetricsCategory.CPU_CATEGORY, MetricsCounter.PROCESSOR_PERCENT_TIME, "test",
                                        10)]
                        configurator.check_cgroups(tracked_metrics)
                        if method_to_fail == "_check_processes_in_agent_cgroup":
                            self.assertFalse(configurator.enabled(), "An error in {0} should have disabled cgroups".format(method_to_fail))
                        else:
                            self.assertFalse(configurator.agent_enabled(), "An error in {0} should have disabled cgroups".format(method_to_fail))

                        disable_events = [kwargs for _, kwargs in add_event.call_args_list if kwargs["op"] == WALAEventOperation.CGroupsDisabled]
                        self.assertTrue(
                            len(disable_events) == 1,
                            "Exactly 1 event should have been emitted when {0} fails. Got: {1}".format(method_to_fail, disable_events))
                        self.assertIn(
                            "[CGroupsException] {0}".format(method_to_fail),
                            disable_events[0]["message"],
                            "The error message is not correct when {0} failed".format(method_to_fail))
                finally:
                    for p in patchers:
                        p.stop()

    def test_check_agent_memory_usage_should_raise_a_cgroups_exception_when_the_limit_is_exceeded(self):
        metrics = [MetricValue(MetricsCategory.MEMORY_CATEGORY, MetricsCounter.TOTAL_MEM_USAGE, AGENT_NAME_TELEMETRY, conf.get_agent_memory_quota() + 1),
                   MetricValue(MetricsCategory.MEMORY_CATEGORY, MetricsCounter.SWAP_MEM_USAGE, AGENT_NAME_TELEMETRY, conf.get_agent_memory_quota() + 1)]

        with self.assertRaises(AgentMemoryExceededException) as context_manager:
            with self._get_cgroup_configurator() as configurator:
                with patch("azurelinuxagent.common.cgroup.MemoryCgroup.get_tracked_metrics") as tracked_metrics:
                    tracked_metrics.return_value = metrics
                    configurator.check_agent_memory_usage()

        self.assertIn("The agent memory limit {0} bytes exceeded".format(conf.get_agent_memory_quota()), ustr(context_manager.exception), "An incorrect exception was raised")


#
# Copyright 2021 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Requires Python 2.6+ and Openssl 1.0+
#

import azurelinuxagent.common.utils.shellutil as shellutil
from azurelinuxagent.common.osutil.default import DefaultOSUtil


class PhotonOSUtil(DefaultOSUtil):

    def __init__(self):
        super(PhotonOSUtil, self).__init__()
        self.agent_conf_file_path = '/etc/waagent.conf'

    @staticmethod
    def get_systemd_unit_file_install_path():
        return '/usr/lib/systemd/system'

    @staticmethod
    def get_agent_bin_path():
        return '/usr/bin'

    def is_dhcp_enabled(self):
        return True

    def start_network(self) :
        return shellutil.run('systemctl start systemd-networkd', chk_err=False)

    def restart_if(self, ifname=None, retries=None, wait=None):
        shellutil.run('systemctl restart systemd-networkd')

    def restart_ssh_service(self):
        shellutil.run('systemctl restart sshd')

    def stop_dhcp_service(self):
        return shellutil.run('systemctl stop systemd-networkd', chk_err=False)

    def start_dhcp_service(self):
        return shellutil.run('systemctl start systemd-networkd', chk_err=False)

    def start_agent_service(self):
        return shellutil.run('systemctl start waagent', chk_err=False)

    def stop_agent_service(self):
        return shellutil.run('systemctl stop waagent', chk_err=False)

    def get_dhcp_pid(self):
        return self._get_dhcp_pid(['pidof', 'systemd-networkd'])

    def conf_sshd(self, disable_password):
        pass



import contextlib
import json
import os.path
import re
import subprocess
import uuid

from azurelinuxagent.common import conf
from azurelinuxagent.common.event import WALAEventOperation
from azurelinuxagent.common.exception import GoalStateAggregateStatusCodes
from azurelinuxagent.common.future import ustr
from azurelinuxagent.common.protocol.restapi import ExtensionRequestedState, ExtensionState
from azurelinuxagent.common.utils import fileutil
from azurelinuxagent.ga.exthandlers import get_exthandlers_handler, ExtensionStatusValue, ExtCommandEnvVariable, \
    GoalStateStatus, ExtHandlerInstance
from tests.ga.extension_emulator import enable_invocations, extension_emulator, ExtensionCommandNames, Actions, \
    extract_extension_info_from_command
from tests.protocol.mocks import mock_wire_protocol, MockHttpResponse
from tests.protocol.HttpRequestPredicates import HttpRequestPredicates
from tests.protocol.mockwiredata import DATA_FILE, WireProtocolData
from tests.tools import AgentTestCase, mock_sleep, patch


class TestMultiConfigExtensionsConfigParsing(AgentTestCase):

    _MULTI_CONFIG_TEST_DATA = os.path.join("wire", "multi-config")

    def setUp(self):
        AgentTestCase.setUp(self)
        self.mock_sleep = patch("time.sleep", lambda *_: mock_sleep(0.0001))
        self.mock_sleep.start()
        self.test_data = DATA_FILE.copy()

    def tearDown(self):
        self.mock_sleep.stop()
        AgentTestCase.tearDown(self)

    class _TestExtHandlerObject:
        def __init__(self, name, version, state="enabled"):
            self.name = name
            self.version = version
            self.state = state
            self.is_invalid_setting = False
            self.settings = dict()

    class _TestExtensionObject:
        def __init__(self, name, seq_no, dependency_level="0", state="enabled"):
            self.name = name
            self.seq_no = seq_no
            self.dependency_level = int(dependency_level)
            self.state = state

    def _mock_and_assert_ext_handlers(self, expected_handlers):
        with mock_wire_protocol(self.test_data) as protocol:
            ext_handlers = protocol.get_goal_state().extensions_goal_state.extensions
            for ext_handler in ext_handlers:
                if ext_handler.name not in expected_handlers:
                    continue
                expected_handler = expected_handlers.pop(ext_handler.name)
                self.assertEqual(expected_handler.state, ext_handler.state)
                self.assertEqual(expected_handler.version, ext_handler.version)
                self.assertEqual(expected_handler.is_invalid_setting, ext_handler.is_invalid_setting)
                self.assertEqual(len(expected_handler.settings), len(ext_handler.settings))

                for extension in ext_handler.settings:
                    self.assertIn(extension.name, expected_handler.settings)
                    expected_extension = expected_handler.settings.pop(extension.name)
                    self.assertEqual(expected_extension.seq_no, extension.sequenceNumber)
                    self.assertEqual(expected_extension.state, extension.state)
                    self.assertEqual(expected_extension.dependency_level, extension.dependencyLevel)

                self.assertEqual(0, len(expected_handler.settings), "All extensions not verified for handler")

            self.assertEqual(0, len(expected_handlers), "All handlers not verified")

    def _get_mock_expected_handler_data(self, rc_extensions, vmaccess_extensions, geneva_extensions):
        # Set expected handler data
        run_command_test_handler = self._TestExtHandlerObject("Microsoft.CPlat.Core.RunCommandHandlerWindows", "2.3.0")
        run_command_test_handler.settings.update(rc_extensions)

        vm_access_test_handler = self._TestExtHandlerObject("Microsoft.Compute.VMAccessAgent", "2.4.7")
        vm_access_test_handler.settings.update(vmaccess_extensions)

        geneva_test_handler = self._TestExtHandlerObject("Microsoft.Azure.Geneva.GenevaMonitoring", "2.20.0.1")
        geneva_test_handler.settings.update(geneva_extensions)

        expected_handlers = {
            run_command_test_handler.name: run_command_test_handler,
            vm_access_test_handler.name: vm_access_test_handler,
            geneva_test_handler.name: geneva_test_handler
        }
        return expected_handlers

    def test_it_should_parse_multi_config_settings_properly(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA, "ext_conf_with_multi_config.xml")

        rc_extensions = dict()
        rc_extensions["firstRunCommand"] = self._TestExtensionObject(name="firstRunCommand", seq_no=2)
        rc_extensions["secondRunCommand"] = self._TestExtensionObject(name="secondRunCommand", seq_no=2,
                                                                      dependency_level="3")
        rc_extensions["thirdRunCommand"] = self._TestExtensionObject(name="thirdRunCommand", seq_no=1,
                                                                     dependency_level="4")

        vmaccess_extensions = {
            "Microsoft.Compute.VMAccessAgent": self._TestExtensionObject(name="Microsoft.Compute.VMAccessAgent",
                                                                         seq_no=1, dependency_level=2)}

        geneva_extensions = {"Microsoft.Azure.Geneva.GenevaMonitoring": self._TestExtensionObject(
            name="Microsoft.Azure.Geneva.GenevaMonitoring", seq_no=1)}

        expected_handlers = self._get_mock_expected_handler_data(rc_extensions, vmaccess_extensions, geneva_extensions)
        self._mock_and_assert_ext_handlers(expected_handlers)

    def test_it_should_parse_multi_config_with_disable_state_properly(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_with_disabled_multi_config.xml")

        rc_extensions = dict()
        rc_extensions["firstRunCommand"] = self._TestExtensionObject(name="firstRunCommand", seq_no=3)
        rc_extensions["secondRunCommand"] = self._TestExtensionObject(name="secondRunCommand", seq_no=3,
                                                                      dependency_level="1")
        rc_extensions["thirdRunCommand"] = self._TestExtensionObject(name="thirdRunCommand", seq_no=1,
                                                                     dependency_level="4", state="disabled")

        vmaccess_extensions = {
            "Microsoft.Compute.VMAccessAgent": self._TestExtensionObject(name="Microsoft.Compute.VMAccessAgent",
                                                                         seq_no=2, dependency_level="2")}

        geneva_extensions = {"Microsoft.Azure.Geneva.GenevaMonitoring": self._TestExtensionObject(
            name="Microsoft.Azure.Geneva.GenevaMonitoring", seq_no=2)}

        expected_handlers = self._get_mock_expected_handler_data(rc_extensions, vmaccess_extensions, geneva_extensions)
        self._mock_and_assert_ext_handlers(expected_handlers)


class _MultiConfigBaseTestClass(AgentTestCase):
    _MULTI_CONFIG_TEST_DATA = os.path.join("wire", "multi-config")

    def setUp(self):
        AgentTestCase.setUp(self)
        self.mock_sleep = patch("time.sleep", lambda *_: mock_sleep(0.01))
        self.mock_sleep.start()
        self.test_data = DATA_FILE.copy()

    def tearDown(self):
        self.mock_sleep.stop()
        AgentTestCase.tearDown(self)

    @contextlib.contextmanager
    def _setup_test_env(self, mock_manifest=False):

        with mock_wire_protocol(self.test_data) as protocol:
            def mock_http_put(url, *args, **_):
                if HttpRequestPredicates.is_host_plugin_status_request(url):
                    # Skip reading the HostGA request data as its encoded
                    return MockHttpResponse(status=500)
                protocol.aggregate_status = json.loads(args[0])
                return MockHttpResponse(status=201)

            with patch("azurelinuxagent.common.agent_supported_feature._MultiConfigFeature.is_supported", True):
                protocol.aggregate_status = None
                protocol.set_http_handlers(http_put_handler=mock_http_put)
                exthandlers_handler = get_exthandlers_handler(protocol)
                no_of_extensions = protocol.mock_wire_data.get_no_of_extensions_in_config()

                if mock_manifest:
                    with patch('azurelinuxagent.ga.exthandlers.HandlerManifest.supports_multiple_extensions',
                               return_value=True):
                        yield exthandlers_handler, protocol, no_of_extensions
                else:
                    yield exthandlers_handler, protocol, no_of_extensions

    def _assert_and_get_handler_status(self, aggregate_status, handler_name="OSTCExtensions.ExampleHandlerLinux",
                                       handler_version="1.0.0", status="Ready", expected_count=1, message=None):
        self.assertIsNotNone(aggregate_status['aggregateStatus'], "No aggregate status found")
        handlers = [handler for handler in aggregate_status['aggregateStatus']['handlerAggregateStatus'] if
                    handler_name == handler['handlerName'] and handler_version == handler['handlerVersion']]
        self.assertEqual(expected_count, len(handlers), "Unexpected extension count")
        self.assertTrue(all(handler['status'] == status for handler in handlers),
                        "Unexpected Status reported for handler {0}".format(handler_name))
        if message is not None:
            self.assertTrue(all(message in handler['formattedMessage']['message'] for handler in handlers),
                            "Status Message mismatch")
        return handlers

    def _assert_extension_status(self, handler_statuses, expected_ext_status, multi_config=False):
        for ext_name, settings_status in expected_ext_status.items():
            ext_status = next(handler for handler in handler_statuses if
                              handler['runtimeSettingsStatus']['settingsStatus']['status']['name'] == ext_name)
            ext_runtime_status = ext_status['runtimeSettingsStatus']
            self.assertIsNotNone(ext_runtime_status, "Extension not found")
            self.assertEqual(settings_status['seq_no'], ext_runtime_status['sequenceNumber'], "Sequence no mismatch")
            self.assertEqual(settings_status['status'], ext_runtime_status['settingsStatus']['status']['status'],
                             "status mismatch")

            if 'message' in settings_status and settings_status['message'] is not None:
                self.assertIn(settings_status['message'],
                              ext_runtime_status['settingsStatus']['status']['formattedMessage']['message'],
                              "message mismatch")

            if multi_config:
                self.assertEqual(ext_name, ext_runtime_status['extensionName'], "ext name mismatch")
            else:
                self.assertNotIn('extensionName', ext_runtime_status, "Extension name should not be reported for SC")

            handler_statuses.remove(ext_status)

        self.assertEqual(0, len(handler_statuses), "Unexpected extensions left for handler")


class TestMultiConfigExtensions(_MultiConfigBaseTestClass):

    def __assert_extension_not_present(self, handlers, extensions):
        for ext_name in extensions:
            self.assertFalse(all(
                'runtimeSettingsStatus' in handler and 'extensionName' in handler['runtimeSettingsStatus']
                and handler['runtimeSettingsStatus']['extensionName'] == ext_name for
                handler in handlers), "Extension status found")

    def __run_and_assert_generic_case(self, exthandlers_handler, protocol, no_of_extensions, with_message=True):

        def get_message(msg):
            return msg if with_message else None

        exthandlers_handler.run()
        exthandlers_handler.report_ext_handlers_status()
        self.assertEqual(no_of_extensions,
                         len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                         "incorrect extensions reported")
        mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                          handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                          expected_count=3)
        expected_extensions = {
            "firstExtension": {"status": ExtensionStatusValue.success, "seq_no": 1,
                               "message": get_message("Enabling firstExtension")},
            "secondExtension": {"status": ExtensionStatusValue.success, "seq_no": 2,
                                "message": get_message("Enabling secondExtension")},
            "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 3,
                               "message": get_message("Enabling thirdExtension")},
        }
        self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

        sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                         handler_name="Microsoft.Powershell.ExampleExtension")
        expected_extensions = {
            "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 9,
                                                      "message": get_message("Enabling SingleConfig extension")}
        }
        self._assert_extension_status(sc_handler[:], expected_extensions)
        return mc_handlers, sc_handler

    def __setup_and_assert_disable_scenario(self, exthandlers_handler, protocol):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA, 'ext_conf_mc_disabled_extensions.xml')
        protocol.mock_wire_data = WireProtocolData(self.test_data)
        protocol.mock_wire_data.set_incarnation(2)
        protocol.client.update_goal_state()
        exthandlers_handler.run()
        exthandlers_handler.report_ext_handlers_status()

        mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                          handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                          status="Ready", expected_count=2)
        expected_extensions = {
            "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 99, "message": None},
            "fourthExtension": {"status": ExtensionStatusValue.success, "seq_no": 101, "message": None},
        }
        self.__assert_extension_not_present(mc_handlers[:], ["firstExtension", "secondExtension"])
        self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)
        sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                         handler_name="Microsoft.Powershell.ExampleExtension",
                                                         status="Ready")
        expected_extensions = {
            "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 10,
                                                      "message": None}
        }
        self._assert_extension_status(sc_handler[:], expected_extensions)
        return mc_handlers, sc_handler

    @contextlib.contextmanager
    def __setup_generic_test_env(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")

        first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension")
        second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension")
        third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension")
        fourth_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension")

        with self._setup_test_env(mock_manifest=True) as (exthandlers_handler, protocol, no_of_extensions):
            with enable_invocations(first_ext, second_ext, third_ext, fourth_ext) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")
                invocation_record.compare(
                    (first_ext, ExtensionCommandNames.INSTALL),
                    (first_ext, ExtensionCommandNames.ENABLE),
                    (second_ext, ExtensionCommandNames.ENABLE),
                    (third_ext, ExtensionCommandNames.ENABLE),
                    (fourth_ext, ExtensionCommandNames.INSTALL),
                    (fourth_ext, ExtensionCommandNames.ENABLE)
                )

            self.__run_and_assert_generic_case(exthandlers_handler, protocol, no_of_extensions, with_message=False)
            yield exthandlers_handler, protocol, [first_ext, second_ext, third_ext, fourth_ext]

    def test_it_should_execute_and_report_multi_config_extensions_properly(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env(mock_manifest=True) as (exthandlers_handler, protocol, no_of_extensions):

            # Case 1: Install and enable Single and MultiConfig extensions
            self.__run_and_assert_generic_case(exthandlers_handler, protocol, no_of_extensions)

            # Case 2: Disable 2 multi-config extensions and add another for enable
            self.__setup_and_assert_disable_scenario(exthandlers_handler, protocol)

            # Case 3: Uninstall Multi-config handler (with enabled extensions) and single config extension
            protocol.mock_wire_data.set_incarnation(3)
            protocol.mock_wire_data.set_extensions_config_state(ExtensionRequestedState.Uninstall)
            protocol.client.update_goal_state()
            exthandlers_handler.run()
            exthandlers_handler.report_ext_handlers_status()
            self.assertEqual(0, len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                             "No handler/extension status should be reported")

    def test_it_should_report_unregistered_version_error_per_extension(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env() as (exthandlers_handler, protocol, no_of_extensions):
            # Set a random failing extension
            failing_version = "19.12.1221"
            protocol.mock_wire_data.set_extensions_config_version(failing_version)
            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()
            exthandlers_handler.run()
            exthandlers_handler.report_ext_handlers_status()
            self.assertEqual(no_of_extensions,
                             len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                             "incorrect extensions reported")
            error_msg_format = '[ExtensionError] Unable to find version {0} in manifest for extension {1}'
            mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                              handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                              handler_version=failing_version, status="NotReady",
                                                              expected_count=3,
                                                              message=error_msg_format.format(failing_version,
                                                                                               "OSTCExtensions.ExampleHandlerLinux"))
            self.assertTrue(all(
                handler['runtimeSettingsStatus']['settingsStatus']['status']['operation'] == WALAEventOperation.Download and
                handler['runtimeSettingsStatus']['settingsStatus']['status']['status'] == ExtensionStatusValue.error for
                handler in mc_handlers), "Incorrect data reported")
            sc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                              handler_name="Microsoft.Powershell.ExampleExtension",
                                                              handler_version=failing_version, status="NotReady",
                                                              message=error_msg_format.format(failing_version,
                                                                                               "Microsoft.Powershell.ExampleExtension"))
            self.assertFalse(all("runtimeSettingsStatus" in handler for handler in sc_handlers), "Incorrect status")

    def test_it_should_not_install_handler_again_if_installed(self):

        with self.__setup_generic_test_env() as (_, _, _):
            # Everything is already asserted in the context manager
            pass

    def test_it_should_retry_handler_installation_per_extension_if_failed(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env() as (exthandlers_handler, protocol, no_of_extensions):
            fail_code, fail_action = Actions.generate_unique_fail()
            first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                           install_action=fail_action, supports_multiple_extensions=True)
            second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                            supports_multiple_extensions=True)
            third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension",
                                           supports_multiple_extensions=True)
            sc_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension", install_action=fail_action)
            with enable_invocations(first_ext, second_ext, third_ext, sc_ext) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")
                invocation_record.compare(
                    (first_ext, ExtensionCommandNames.INSTALL),
                    # Should try installation again if first time failed
                    (second_ext, ExtensionCommandNames.INSTALL),
                    (second_ext, ExtensionCommandNames.ENABLE),
                    (third_ext, ExtensionCommandNames.ENABLE),
                    (sc_ext, ExtensionCommandNames.INSTALL)
                )
                mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                  expected_count=3, status="Ready")
                expected_extensions = {
                    "firstExtension": {"status": ExtensionStatusValue.error, "seq_no": 1, "message": fail_code},
                    "secondExtension": {"status": ExtensionStatusValue.success, "seq_no": 2, "message": None},
                    "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 3, "message": None},
                }
                self._assert_extension_status(mc_handlers, expected_extensions, multi_config=True)

                sc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="Microsoft.Powershell.ExampleExtension",
                                                                  status="NotReady", message=fail_code)
                self.assertFalse(all("runtimeSettingsStatus" in handler for handler in sc_handlers), "Incorrect status")

    def test_it_should_only_disable_enabled_extensions_on_update(self):
        with self.__setup_generic_test_env() as (exthandlers_handler, protocol, old_exts):

            # Update extensions
            self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA, 'ext_conf_mc_update_extensions.xml')
            protocol.mock_wire_data = WireProtocolData(self.test_data)
            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()

            new_version = "1.1.0"
            new_first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                               version=new_version, supports_multiple_extensions=True)
            new_second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                                version=new_version, supports_multiple_extensions=True)
            new_third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension",
                                               version=new_version, supports_multiple_extensions=True)
            new_sc_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension", version=new_version)
            with enable_invocations(new_first_ext, new_second_ext, new_third_ext, new_sc_ext, *old_exts) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                old_first, old_second, old_third, old_fourth = old_exts
                invocation_record.compare(
                    # Disable all enabled commands for MC before updating the Handler
                    (old_first, ExtensionCommandNames.DISABLE),
                    (old_second, ExtensionCommandNames.DISABLE),
                    (old_third, ExtensionCommandNames.DISABLE),
                    (new_first_ext, ExtensionCommandNames.UPDATE),
                    (old_first, ExtensionCommandNames.UNINSTALL),
                    (new_first_ext, ExtensionCommandNames.INSTALL),
                    # No enable for First and Second extension as their state is Disabled in GoalState,
                    # only enabled the ThirdExtension
                    (new_third_ext, ExtensionCommandNames.ENABLE),
                    # Follow the normal update pattern for Single config handlers
                    (old_fourth, ExtensionCommandNames.DISABLE),
                    (new_sc_ext, ExtensionCommandNames.UPDATE),
                    (old_fourth, ExtensionCommandNames.UNINSTALL),
                    (new_sc_ext, ExtensionCommandNames.INSTALL),
                    (new_sc_ext, ExtensionCommandNames.ENABLE)
                )

            mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                              handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                              expected_count=1, handler_version=new_version)
            expected_extensions = {
                "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 99, "message": None}
            }
            self._assert_extension_status(mc_handlers, expected_extensions, multi_config=True)
            self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status, handler_version=new_version,
                                                handler_name="Microsoft.Powershell.ExampleExtension")

    def test_it_should_retry_update_sequence_per_extension_if_previous_failed(self):
        with self.__setup_generic_test_env() as (exthandlers_handler, protocol, old_exts):
            # Update extensions
            self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA, 'ext_conf_mc_update_extensions.xml')
            protocol.mock_wire_data = WireProtocolData(self.test_data)
            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()

            new_version = "1.1.0"
            _, fail_action = Actions.generate_unique_fail()
            # Fail Uninstall of the secondExtension
            old_exts[1] = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                             uninstall_action=fail_action, supports_multiple_extensions=True)
            # Fail update of the first extension
            new_first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                               version=new_version, update_action=fail_action,
                                               supports_multiple_extensions=True)
            new_second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                                version=new_version, supports_multiple_extensions=True)
            new_third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension",
                                               version=new_version, supports_multiple_extensions=True)
            new_sc_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension", version=new_version)

            with enable_invocations(new_first_ext, new_second_ext, new_third_ext, new_sc_ext,
                                    *old_exts) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                old_first, old_second, old_third, old_fourth = old_exts
                invocation_record.compare(
                    # Disable all enabled commands for MC before updating the Handler
                    (old_first, ExtensionCommandNames.DISABLE),
                    (old_second, ExtensionCommandNames.DISABLE),
                    (old_third, ExtensionCommandNames.DISABLE),
                    (new_first_ext, ExtensionCommandNames.UPDATE),
                    # Since the extensions have been disabled before, we won't disable them again for Update scenario
                    (new_second_ext, ExtensionCommandNames.UPDATE),
                    # This will fail too as per the mock above
                    (old_second, ExtensionCommandNames.UNINSTALL),
                    (new_third_ext, ExtensionCommandNames.UPDATE),
                    (old_third, ExtensionCommandNames.UNINSTALL),
                    (new_third_ext, ExtensionCommandNames.INSTALL),
                    # No enable for First and Second extension as their state is Disabled in GoalState,
                    # only enabled the ThirdExtension
                    (new_third_ext, ExtensionCommandNames.ENABLE),
                    # Follow the normal update pattern for Single config handlers
                    (old_fourth, ExtensionCommandNames.DISABLE),
                    (new_sc_ext, ExtensionCommandNames.UPDATE),
                    (old_fourth, ExtensionCommandNames.UNINSTALL),
                    (new_sc_ext, ExtensionCommandNames.INSTALL),
                    (new_sc_ext, ExtensionCommandNames.ENABLE)
                )

            # Since firstExtension and secondExtension are Disabled, we won't report their status
            mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                              handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                              expected_count=1, handler_version=new_version)
            expected_extensions = {
                "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 99, "message": None}
            }
            self._assert_extension_status(mc_handlers, expected_extensions, multi_config=True)
            sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                             handler_version=new_version,
                                                             handler_name="Microsoft.Powershell.ExampleExtension")
            expected_extensions = {
                "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 10,
                                                          "message": None}
            }
            self._assert_extension_status(sc_handler, expected_extensions)

    def test_it_should_report_disabled_extension_errors_if_update_failed(self):
        with self.__setup_generic_test_env() as (exthandlers_handler, protocol, old_exts):
            # Update extensions
            self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA, 'ext_conf_mc_update_extensions.xml')
            protocol.mock_wire_data = WireProtocolData(self.test_data)
            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()

            new_version = "1.1.0"
            fail_code, fail_action = Actions.generate_unique_fail()
            # Fail Disable of the firstExtension
            old_exts[0] = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                             disable_action=fail_action, supports_multiple_extensions=True)
            new_first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                               version=new_version, supports_multiple_extensions=True)
            new_second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                                version=new_version, supports_multiple_extensions=True)
            new_third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension",
                                               version=new_version, supports_multiple_extensions=True)
            new_fourth_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension", version=new_version)

            with enable_invocations(new_first_ext, new_second_ext, new_third_ext, new_fourth_ext,
                                    *old_exts) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                old_first, _, _, old_fourth = old_exts
                invocation_record.compare(
                    # Disable for firstExtension should fail 3 times, i.e., once per extension which tries to update the Handler
                    (old_first, ExtensionCommandNames.DISABLE),
                    (old_first, ExtensionCommandNames.DISABLE),
                    (old_first, ExtensionCommandNames.DISABLE),
                    # Since Disable fails for the firstExtension and continueOnUpdate = False, Update should not go through
                    # Follow the normal update pattern for Single config handlers
                    (old_fourth, ExtensionCommandNames.DISABLE),
                    (new_fourth_ext, ExtensionCommandNames.UPDATE),
                    (old_fourth, ExtensionCommandNames.UNINSTALL),
                    (new_fourth_ext, ExtensionCommandNames.INSTALL),
                    (new_fourth_ext, ExtensionCommandNames.ENABLE)
                )

            # Since firstExtension and secondExtension are Disabled, we won't report their status
            mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                              handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                              expected_count=1, handler_version=new_version,
                                                              status="NotReady", message=fail_code)
            expected_extensions = {
                "thirdExtension": {"status": ExtensionStatusValue.error, "seq_no": 99, "message": fail_code}
            }
            self._assert_extension_status(mc_handlers, expected_extensions, multi_config=True)
            sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                             handler_version=new_version,
                                                             handler_name="Microsoft.Powershell.ExampleExtension")
            expected_extensions = {
                "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 10,
                                                          "message": None}
            }
            self._assert_extension_status(sc_handler, expected_extensions)

    def test_it_should_report_extension_status_properly(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env(mock_manifest=True) as (exthandlers_handler, protocol, no_of_extensions):
            self.__run_and_assert_generic_case(exthandlers_handler, protocol, no_of_extensions)

    def test_it_should_handle_and_report_enable_errors_properly(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env() as (exthandlers_handler, protocol, no_of_extensions):
            fail_code, fail_action = Actions.generate_unique_fail()
            first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                           supports_multiple_extensions=True)
            second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                            supports_multiple_extensions=True)
            third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension",
                                           enable_action=fail_action, supports_multiple_extensions=True)
            fourth_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension", enable_action=fail_action)
            with enable_invocations(first_ext, second_ext, third_ext, fourth_ext) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")
                invocation_record.compare(
                    (first_ext, ExtensionCommandNames.INSTALL),
                    (first_ext, ExtensionCommandNames.ENABLE),
                    (second_ext, ExtensionCommandNames.ENABLE),
                    (third_ext, ExtensionCommandNames.ENABLE),
                    (fourth_ext, ExtensionCommandNames.INSTALL),
                    (fourth_ext, ExtensionCommandNames.ENABLE)
                )
                mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                  expected_count=3, status="Ready")
                expected_extensions = {
                    "firstExtension": {"status": ExtensionStatusValue.success, "seq_no": 1, "message": None},
                    "secondExtension": {"status": ExtensionStatusValue.success, "seq_no": 2, "message": None},
                    "thirdExtension": {"status": ExtensionStatusValue.error, "seq_no": 3, "message": fail_code},
                }
                self._assert_extension_status(mc_handlers, expected_extensions, multi_config=True)

                sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                 handler_name="Microsoft.Powershell.ExampleExtension",
                                                                 status="NotReady", message=fail_code)
                expected_extensions = {
                    "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.error, "seq_no": 9,
                                                              "message": fail_code}
                }
                self._assert_extension_status(sc_handler, expected_extensions)

    def test_it_should_cleanup_extension_state_on_disable(self):

        def __assert_state_file(handler_name, handler_version, extensions, state, not_present=None):
            config_path = os.path.join(self.tmp_dir, "{0}-{1}".format(handler_name, handler_version), "config")
            config_files = os.listdir(config_path)

            for ext_name in extensions:
                self.assertIn("{0}.settings".format(ext_name), config_files, "settings not found")
                self.assertEqual(
                    fileutil.read_file(os.path.join(config_path, "{0}.HandlerState".format(ext_name.split(".")[0]))),
                    state, "Invalid state")

            if not_present is not None:
                for ext_name in not_present:
                    self.assertNotIn("{0}.HandlerState".format(ext_name), config_files, "Wrongful state found")

        with self.__setup_generic_test_env() as (ext_handler, protocol, _):
            __assert_state_file("OSTCExtensions.ExampleHandlerLinux", "1.0.0",
                                ["firstExtension.1", "secondExtension.2", "thirdExtension.3"], ExtensionState.Enabled)

            self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA, 'ext_conf_mc_disabled_extensions.xml')
            protocol.mock_wire_data = WireProtocolData(self.test_data)
            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()

            ext_handler.run()
            ext_handler.report_ext_handlers_status()
            mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                              handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                              expected_count=2, status="Ready")
            expected_extensions = {
                "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 99, "message": "Enabling thirdExtension"},
                "fourthExtension": {"status": ExtensionStatusValue.success, "seq_no": 101, "message": "Enabling fourthExtension"},
            }
            self._assert_extension_status(mc_handlers, expected_extensions, multi_config=True)
            __assert_state_file("OSTCExtensions.ExampleHandlerLinux", "1.0.0",
                                ["thirdExtension.99", "fourthExtension.101"], ExtensionState.Enabled,
                                not_present=["firstExtension", "secondExtension"])

            sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                             handler_name="Microsoft.Powershell.ExampleExtension")
            expected_extensions = {
                "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 10,
                                                          "message": "Enabling SingleConfig Extension"}
            }
            self._assert_extension_status(sc_handler, expected_extensions)

    def test_it_should_create_command_execution_log_per_extension(self):
        with self.__setup_generic_test_env() as (_, _, _):
            sc_handler_path = os.path.join(conf.get_ext_log_dir(), "Microsoft.Powershell.ExampleExtension")
            mc_handler_path = os.path.join(conf.get_ext_log_dir(), "OSTCExtensions.ExampleHandlerLinux")
            self.assertIn("CommandExecution_firstExtension.log", os.listdir(mc_handler_path),
                          "Command Execution file not found")
            self.assertGreater(os.path.getsize(os.path.join(mc_handler_path, "CommandExecution_firstExtension.log")), 0,
                               "Log file not being used")
            self.assertIn("CommandExecution_secondExtension.log", os.listdir(mc_handler_path),
                          "Command Execution file not found")
            self.assertGreater(os.path.getsize(os.path.join(mc_handler_path, "CommandExecution_secondExtension.log")), 0,
                               "Log file not being used")
            self.assertIn("CommandExecution_thirdExtension.log", os.listdir(mc_handler_path),
                          "Command Execution file not found")
            self.assertGreater(os.path.getsize(os.path.join(mc_handler_path, "CommandExecution_thirdExtension.log")), 0,
                               "Log file not being used")
            self.assertIn("CommandExecution.log", os.listdir(sc_handler_path), "Command Execution file not found")
            self.assertGreater(os.path.getsize(os.path.join(sc_handler_path, "CommandExecution.log")), 0,
                               "Log file not being used")

    def test_it_should_set_relevant_environment_variables_for_mc(self):
        original_popen = subprocess.Popen
        handler_envs = {}

        def __assert_env_variables(handler_name, handler_version="1.0.0", seq_no="1", ext_name=None, expected_vars=None,
                                   not_expected=None):
            original_env_vars = {
                ExtCommandEnvVariable.ExtensionPath: os.path.join(self.tmp_dir, "{0}-{1}".format(handler_name, handler_version)),
                ExtCommandEnvVariable.ExtensionVersion: handler_version,
                ExtCommandEnvVariable.ExtensionSeqNumber: ustr(seq_no),
                ExtCommandEnvVariable.WireProtocolAddress: '168.63.129.16',
                ExtCommandEnvVariable.ExtensionSupportedFeatures: json.dumps([{"Key": "ExtensionTelemetryPipeline",
                                                                               "Value": "1.0"}])

            }

            full_name = handler_name
            if ext_name is not None:
                original_env_vars[ExtCommandEnvVariable.ExtensionName] = ext_name
                full_name = "{0}.{1}".format(handler_name, ext_name)

            self.assertIn(full_name, handler_envs, "Handler/ext combo not called")
            for commands in handler_envs[full_name]:
                expected_environment_variables = original_env_vars.copy()
                if expected_vars is not None and commands['command'] in expected_vars:
                    for name, val in expected_vars[commands['command']].items():
                        expected_environment_variables[name] = val

                self.assertTrue(all(
                    env_var in commands['data'] and env_val == commands['data'][env_var] for env_var, env_val in
                    expected_environment_variables.items()),
                    "Incorrect data for environment variable for {0}-{1}, incorrect: {2}".format(
                        full_name, commands['command'],
                        [(env_var, env_val) for env_var, env_val in expected_environment_variables.items() if
                         env_var not in commands['data'] or env_val != commands['data'][env_var]]))

                if not_expected is not None and commands['command'] in not_expected:
                    self.assertFalse(any(env_var in commands['data'] for env_var in not_expected), "Unwanted env variable found")

        def mock_popen(cmd, *_, **kwargs):
            # This cgroupsapi Popen mocking all other popen calls which breaking the extension emulator logic.
            # The emulator should be used only on extension commands and not on other commands even env flag set.
            # So, added ExtensionVersion check to avoid using extension emulator on non extension operations.
            if 'env' in kwargs and ExtCommandEnvVariable.ExtensionVersion in kwargs['env']:
                handler_name, __, command = extract_extension_info_from_command(cmd)
                name = handler_name
                if ExtCommandEnvVariable.ExtensionName in kwargs['env']:
                    name = "{0}.{1}".format(handler_name, kwargs['env'][ExtCommandEnvVariable.ExtensionName])

                data = {
                    "command": command,
                    "data": kwargs['env']
                }
                if name in handler_envs:
                    handler_envs[name].append(data)
                else:
                    handler_envs[name] = [data]
            return original_popen(cmd, *_, **kwargs)

        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env(mock_manifest=True) as (exthandlers_handler, protocol, no_of_extensions):
            with patch('azurelinuxagent.common.cgroupapi.subprocess.Popen', side_effect=mock_popen):
                # Case 1: Check normal scenario - Install/Enable
                mc_handlers, sc_handler = self.__run_and_assert_generic_case(exthandlers_handler, protocol,
                                                                             no_of_extensions)

                for handler in mc_handlers:
                    __assert_env_variables(handler['handlerName'],
                                           ext_name=handler['runtimeSettingsStatus']['extensionName'],
                                           seq_no=handler['runtimeSettingsStatus']['sequenceNumber'])
                for handler in sc_handler:
                    __assert_env_variables(handler['handlerName'],
                                           seq_no=handler['runtimeSettingsStatus']['sequenceNumber'])

                # Case 2: Check Update Scenario
                # Clear old test case state
                handler_envs = {}
                self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                          'ext_conf_mc_update_extensions.xml')
                protocol.mock_wire_data = WireProtocolData(self.test_data)
                protocol.mock_wire_data.set_incarnation(2)
                protocol.client.update_goal_state()
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()

                mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                  expected_count=1, handler_version="1.1.0")
                expected_extensions = {
                    "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 99,
                                       "message": "Enabling thirdExtension"},
                }
                self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

                sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                 handler_name="Microsoft.Powershell.ExampleExtension",
                                                                 handler_version="1.1.0")
                expected_extensions = {
                    "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 10,
                                                              "message": "Enabling SingleConfig extension"}
                }
                self._assert_extension_status(sc_handler[:], expected_extensions)

                for handler in mc_handlers:
                    __assert_env_variables(handler['handlerName'],
                                           handler_version="1.1.0",
                                           ext_name=handler['runtimeSettingsStatus']['extensionName'],
                                           seq_no=handler['runtimeSettingsStatus']['sequenceNumber'],
                                           expected_vars={
                                                "disable": {
                                                    ExtCommandEnvVariable.ExtensionPath: os.path.join(self.tmp_dir, "{0}-{1}".format(handler['handlerName'], "1.0.0")),
                                                    ExtCommandEnvVariable.ExtensionVersion: '1.0.0'
                                                }})

                # Assert the environment variables were present even for disabled/uninstalled commands
                first_ext_expected_vars = {
                    "disable": {
                        ExtCommandEnvVariable.ExtensionPath: os.path.join(self.tmp_dir, "{0}-{1}".format(handler['handlerName'], "1.0.0")),
                        ExtCommandEnvVariable.ExtensionVersion: '1.0.0'
                    },
                    "uninstall": {
                        ExtCommandEnvVariable.ExtensionPath: os.path.join(self.tmp_dir, "{0}-{1}".format(handler['handlerName'], "1.0.0")),
                        ExtCommandEnvVariable.ExtensionVersion: '1.0.0'
                    },
                    "update": {
                        ExtCommandEnvVariable.UpdatingFromVersion: "1.0.0",
                        ExtCommandEnvVariable.DisableReturnCodeMultipleExtensions:
                            json.dumps([
                                {"extensionName": "firstExtension", "exitCode": "0"},
                                {"extensionName": "secondExtension", "exitCode": "0"},
                                {"extensionName": "thirdExtension", "exitCode": "0"}
                            ])
                    }
                }
                __assert_env_variables(handler['handlerName'], ext_name="firstExtension",
                                       expected_vars=first_ext_expected_vars, handler_version="1.1.0", seq_no="1",
                                       not_expected={
                                           "update": [ExtCommandEnvVariable.DisableReturnCode]
                                       })
                __assert_env_variables(handler['handlerName'], ext_name="secondExtension", seq_no="2")

                for handler in sc_handler:
                    sc_expected_vars = {
                        "disable": {
                            ExtCommandEnvVariable.ExtensionPath: os.path.join(self.tmp_dir, "{0}-{1}".format(handler['handlerName'], "1.0.0")),
                            ExtCommandEnvVariable.ExtensionVersion: '1.0.0'
                        },
                        "uninstall": {
                            ExtCommandEnvVariable.ExtensionPath: os.path.join(self.tmp_dir, "{0}-{1}".format(handler['handlerName'], "1.0.0")),
                            ExtCommandEnvVariable.ExtensionVersion: '1.0.0'
                        },
                        "update": {
                            ExtCommandEnvVariable.UpdatingFromVersion: "1.0.0",
                            ExtCommandEnvVariable.DisableReturnCode: "0"
                        }
                    }
                    __assert_env_variables(handler['handlerName'], handler_version="1.1.0",
                                           seq_no=handler['runtimeSettingsStatus']['sequenceNumber'],
                                           expected_vars=sc_expected_vars, not_expected={
                                           "update": [ExtCommandEnvVariable.DisableReturnCodeMultipleExtensions]
                                       })

    def test_it_should_ignore_disable_errors_for_multi_config_extensions(self):
        fail_code, fail_action = Actions.generate_unique_fail()

        with self.__setup_generic_test_env() as (exthandlers_handler, protocol, exts):

            # Fail disable of 1st and 2nd extension
            exts[0] = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                         disable_action=fail_action)
            exts[1] = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension",
                                         disable_action=fail_action)
            fourth_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.fourthExtension")

            with patch.object(ExtHandlerInstance, "report_event", autospec=True) as patch_report_event:
                with enable_invocations(fourth_ext, *exts) as invocation_record:
                    # Assert even though 2 extensions are failing, we clean their state up properly and enable the
                    # remaining extensions
                    self.__setup_and_assert_disable_scenario(exthandlers_handler, protocol)
                    first_ext, second_ext, third_ext, sc_ext = exts
                    invocation_record.compare(
                        (first_ext, ExtensionCommandNames.DISABLE),
                        (second_ext, ExtensionCommandNames.DISABLE),
                        (third_ext, ExtensionCommandNames.ENABLE),
                        (fourth_ext, ExtensionCommandNames.ENABLE),
                        (sc_ext, ExtensionCommandNames.ENABLE)
                    )

                    reported_events = [kwargs for _, kwargs in patch_report_event.call_args_list if
                                       re.search("Executing command: (.+) with environment variables: ",
                                                 kwargs['message']) is None]

                    self.assertTrue(all(
                        fail_code in kwargs['message'] for kwargs in reported_events if
                        kwargs['name'] == first_ext.name), "Error not reported")
                    self.assertTrue(all(
                        fail_code in kwargs['message'] for kwargs in reported_events if
                        kwargs['name'] == second_ext.name), "Error not reported")
                    # Make sure fail code is not reported for any other extension
                    self.assertFalse(all(
                        fail_code in kwargs['message'] for kwargs in reported_events if
                        kwargs['name'] == third_ext.name), "Error not reported")

    def test_it_should_report_transitioning_if_status_file_not_found(self):
        original_popen = subprocess.Popen

        def mock_popen(cmd, *_, **kwargs):
            if 'env' in kwargs:
                handler_name, handler_version, __ = extract_extension_info_from_command(cmd)
                ext_name = None
                if ExtCommandEnvVariable.ExtensionName in kwargs['env']:
                    ext_name = kwargs['env'][ExtCommandEnvVariable.ExtensionName]
                seq_no = kwargs['env'][ExtCommandEnvVariable.ExtensionSeqNumber]
                status_file_name = "{0}.status".format(seq_no)
                status_file_name = "{0}.{1}".format(ext_name, status_file_name) if ext_name is not None else status_file_name
                status_file = os.path.join(self.tmp_dir, "{0}-{1}".format(handler_name, handler_version), "status", status_file_name)
                if os.path.exists(status_file):
                    os.remove(status_file)

            return original_popen("echo " + cmd, *_, **kwargs)

        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")
        with self._setup_test_env(mock_manifest=True) as (exthandlers_handler, protocol, no_of_extensions):
            with patch('azurelinuxagent.common.cgroupapi.subprocess.Popen', side_effect=mock_popen):
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")
                mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                  expected_count=3)
                agent_status_message = "This status is being reported by the Guest Agent since no status file was " \
                                       "reported by extension {0}: " \
                                       "[ExtensionStatusError] Status file"
                expected_extensions = {
                    "firstExtension": {"status": ExtensionStatusValue.transitioning, "seq_no": 1,
                                       "message": agent_status_message.format("OSTCExtensions.ExampleHandlerLinux.firstExtension")},
                    "secondExtension": {"status": ExtensionStatusValue.transitioning, "seq_no": 2,
                                        "message": agent_status_message.format("OSTCExtensions.ExampleHandlerLinux.secondExtension")},
                    "thirdExtension": {"status": ExtensionStatusValue.transitioning, "seq_no": 3,
                                       "message": agent_status_message.format("OSTCExtensions.ExampleHandlerLinux.thirdExtension")},
                }
                self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

                sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                 handler_name="Microsoft.Powershell.ExampleExtension")
                expected_extensions = {
                    "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.transitioning, "seq_no": 9,
                                                              "message": agent_status_message.format("Microsoft.Powershell.ExampleExtension")}
                }
                self._assert_extension_status(sc_handler[:], expected_extensions)

    def test_it_should_report_status_correctly_for_unsupported_goal_state(self):
        with self.__setup_generic_test_env() as (exthandlers_handler, protocol, _):

            # Update GS with an ExtensionConfig with 3 Required features to force GA to mark it as unsupported
            self.test_data['ext_conf'] = "wire/ext_conf_required_features.xml"
            protocol.mock_wire_data = WireProtocolData(self.test_data)
            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()
            # Assert the extension status is the same as we reported for Incarnation 1.
            self.__run_and_assert_generic_case(exthandlers_handler, protocol, no_of_extensions=4, with_message=False)

            # Assert the GS was reported as unsupported
            gs_aggregate_status = protocol.aggregate_status['aggregateStatus']['vmArtifactsAggregateStatus'][
                'goalStateAggregateStatus']
            self.assertEqual(gs_aggregate_status['status'], GoalStateStatus.Failed, "Incorrect status")
            self.assertEqual(gs_aggregate_status['code'],
                             GoalStateAggregateStatusCodes.GoalStateUnsupportedRequiredFeatures, "Incorrect code")
            self.assertEqual(gs_aggregate_status['inSvdSeqNo'], '2', "Incorrect incarnation reported")
            self.assertEqual(gs_aggregate_status['formattedMessage']['message'],
                             'Failing GS incarnation_2 as Unsupported features found: TestRequiredFeature1, TestRequiredFeature2, TestRequiredFeature3',
                             "Incorrect error message reported")

    def test_it_should_fail_handler_if_handler_does_not_support_mc(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_multi_config_no_dependencies.xml")

        first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension")
        second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension")
        third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension")
        fourth_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension")

        with self._setup_test_env() as (exthandlers_handler, protocol, no_of_extensions):
            with enable_invocations(first_ext, second_ext, third_ext, fourth_ext) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")

                invocation_record.compare(
                    # Since we raise a ConfigError, we shouldn't process any of the MC extensions at all
                    (fourth_ext, ExtensionCommandNames.INSTALL),
                    (fourth_ext, ExtensionCommandNames.ENABLE)
                )

                err_msg = 'Handler OSTCExtensions.ExampleHandlerLinux does not support MultiConfig but CRP expects it, failing due to inconsistent data'
                mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                  expected_count=3, status="NotReady", message=err_msg)
                expected_extensions = {
                    "firstExtension": {"status": ExtensionStatusValue.error, "seq_no": 1, "message": err_msg},
                    "secondExtension": {"status": ExtensionStatusValue.error, "seq_no": 2, "message": err_msg},
                    "thirdExtension": {"status": ExtensionStatusValue.error, "seq_no": 3, "message": err_msg},
                }
                self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

                sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                 handler_name="Microsoft.Powershell.ExampleExtension")
                expected_extensions = {
                    "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 9}
                }
                self._assert_extension_status(sc_handler[:], expected_extensions)

    def test_it_should_check_every_time_if_handler_supports_mc(self):
        with self.__setup_generic_test_env() as (exthandlers_handler, protocol, old_exts):

            protocol.mock_wire_data.set_incarnation(2)
            protocol.client.update_goal_state()

            # Mock manifest to not support multiple extensions
            with patch('azurelinuxagent.ga.exthandlers.HandlerManifest.supports_multiple_extensions', return_value=False):
                with enable_invocations(*old_exts) as invocation_record:
                    (_, _, _, fourth_ext) = old_exts
                    exthandlers_handler.run()
                    exthandlers_handler.report_ext_handlers_status()
                    self.assertEqual(4, len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                     "incorrect extensions reported")

                    invocation_record.compare(
                        # Since we raise a ConfigError, we shouldn't process any of the MC extensions at all
                        (fourth_ext, ExtensionCommandNames.ENABLE)
                    )

                    err_msg = 'Handler OSTCExtensions.ExampleHandlerLinux does not support MultiConfig but CRP expects it, failing due to inconsistent data'
                    mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                      handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                      expected_count=3, status="NotReady", message=err_msg)

                    # Since the extensions were not even executed, their status file should reflect the last status
                    # (Handler status above should always report the error though)
                    expected_extensions = {
                        "firstExtension": {"status": ExtensionStatusValue.success, "seq_no": 1},
                        "secondExtension": {"status": ExtensionStatusValue.success, "seq_no": 2},
                        "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 3},
                    }
                    self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

                    sc_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                     handler_name="Microsoft.Powershell.ExampleExtension")
                    expected_extensions = {
                        "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 9}
                    }
                    self._assert_extension_status(sc_handler[:], expected_extensions)


class TestMultiConfigExtensionSequencing(_MultiConfigBaseTestClass):

    @contextlib.contextmanager
    def __setup_test_and_get_exts(self):
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_with_multi_config_dependencies.xml")

        first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension", supports_multiple_extensions=True)
        second_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.secondExtension", supports_multiple_extensions=True)
        third_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.thirdExtension", supports_multiple_extensions=True)
        dependent_sc_ext = extension_emulator(name="Microsoft.Powershell.ExampleExtension")
        independent_sc_ext = extension_emulator(name="Microsoft.Azure.Geneva.GenevaMonitoring", version="1.1.0")

        with self._setup_test_env() as (exthandlers_handler, protocol, no_of_extensions):
            yield exthandlers_handler, protocol, no_of_extensions, first_ext, second_ext, third_ext, dependent_sc_ext, independent_sc_ext

    def test_it_should_process_dependency_chain_extensions_properly(self):
        with self.__setup_test_and_get_exts() as (
            exthandlers_handler, protocol, no_of_extensions, first_ext, second_ext, third_ext, dependent_sc_ext,
                independent_sc_ext):
            with enable_invocations(first_ext, second_ext, third_ext, dependent_sc_ext, independent_sc_ext) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")
                invocation_record.compare(
                    (first_ext, ExtensionCommandNames.INSTALL),
                    (first_ext, ExtensionCommandNames.ENABLE),
                    (independent_sc_ext, ExtensionCommandNames.INSTALL),
                    (independent_sc_ext, ExtensionCommandNames.ENABLE),
                    (dependent_sc_ext, ExtensionCommandNames.INSTALL),
                    (dependent_sc_ext, ExtensionCommandNames.ENABLE),
                    (second_ext, ExtensionCommandNames.ENABLE),
                    (third_ext, ExtensionCommandNames.ENABLE)
                )

                mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                  handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                                  expected_count=3)
                expected_extensions = {
                    "firstExtension": {"status": ExtensionStatusValue.success, "seq_no": 2},
                    "secondExtension": {"status": ExtensionStatusValue.success, "seq_no": 2},
                    "thirdExtension": {"status": ExtensionStatusValue.success, "seq_no": 1},
                }
                self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

                sc_dependent_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                           handler_name="Microsoft.Powershell.ExampleExtension")
                expected_extensions = {
                    "Microsoft.Powershell.ExampleExtension": {"status": ExtensionStatusValue.success, "seq_no": 2}
                }
                self._assert_extension_status(sc_dependent_handler[:], expected_extensions)
                sc_independent_handler = self._assert_and_get_handler_status(
                    aggregate_status=protocol.aggregate_status, handler_name="Microsoft.Azure.Geneva.GenevaMonitoring",
                    handler_version="1.1.0")
                expected_extensions = {
                    "Microsoft.Azure.Geneva.GenevaMonitoring": {"status": ExtensionStatusValue.success, "seq_no": 1}
                }
                self._assert_extension_status(sc_independent_handler[:], expected_extensions)

    def __assert_invalid_status_scenario(self, protocol, fail_code, mc_status="NotReady",
                                         mc_message="Plugin installed but not enabled", err_msg=None):
        mc_handlers = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                          handler_name="OSTCExtensions.ExampleHandlerLinux",
                                                          expected_count=3, status=mc_status,
                                                          message=mc_message)

        expected_extensions = {
            "firstExtension": {"status": ExtensionStatusValue.error, "seq_no": 2, "message": fail_code},
            "secondExtension": {"status": ExtensionStatusValue.error, "seq_no": 2, "message": err_msg},
            "thirdExtension": {"status": ExtensionStatusValue.error, "seq_no": 1, "message": err_msg},
        }
        self._assert_extension_status(mc_handlers[:], expected_extensions, multi_config=True)

        sc_dependent_handler = self._assert_and_get_handler_status(aggregate_status=protocol.aggregate_status,
                                                                   handler_name="Microsoft.Powershell.ExampleExtension",
                                                                   status="NotReady", message=err_msg)
        self.assertTrue(all('runtimeSettingsStatus' not in handler for handler in sc_dependent_handler))

        sc_independent_handler = self._assert_and_get_handler_status(
            aggregate_status=protocol.aggregate_status, handler_name="Microsoft.Azure.Geneva.GenevaMonitoring",
            handler_version="1.1.0", status="NotReady", message=err_msg)
        self.assertTrue(all('runtimeSettingsStatus' not in handler for handler in sc_independent_handler))

    def test_it_should_report_extension_status_failures_for_all_dependent_extensions(self):
        with self.__setup_test_and_get_exts() as (
            exthandlers_handler, protocol, no_of_extensions, first_ext, second_ext, third_ext, dependent_sc_ext,
                independent_sc_ext):

            # Fail the enable for firstExtension.
            fail_code, fail_action = Actions.generate_unique_fail()
            first_ext = extension_emulator(name="OSTCExtensions.ExampleHandlerLinux.firstExtension",
                                           enable_action=fail_action, supports_multiple_extensions=True)

            with enable_invocations(first_ext, second_ext, third_ext, dependent_sc_ext,
                                    independent_sc_ext) as invocation_record:
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()
                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")

                # Since firstExtension is high up on the dependency chain, no other extensions should be executed
                invocation_record.compare(
                    (first_ext, ExtensionCommandNames.INSTALL),
                    (first_ext, ExtensionCommandNames.ENABLE)
                )

                err_msg = 'Skipping processing of extensions since execution of dependent extension OSTCExtensions.ExampleHandlerLinux.firstExtension failed'
                self.__assert_invalid_status_scenario(protocol, fail_code, err_msg=err_msg)

    def test_it_should_stop_execution_if_status_file_contains_errors(self):
        # This test tests the scenario where the extensions exit with a success exit code but fail subsequently with an
        # error in the status file
        self.test_data['ext_conf'] = os.path.join(self._MULTI_CONFIG_TEST_DATA,
                                                  "ext_conf_with_multi_config_dependencies.xml")

        original_popen = subprocess.Popen
        invocation_records = []
        fail_code = str(uuid.uuid4())

        def mock_popen(cmd, *_, **kwargs):

            try:
                handler_name, handler_version, command_name = extract_extension_info_from_command(cmd)
            except ValueError:
                return original_popen(cmd, *_, **kwargs)

            if 'env' in kwargs:
                env = kwargs['env']
                if ExtCommandEnvVariable.ExtensionName in env:
                    full_name = "{0}.{1}".format(handler_name, env[ExtCommandEnvVariable.ExtensionName])
                    status_file = "{0}.{1}.status".format(env[ExtCommandEnvVariable.ExtensionName],
                                                          env[ExtCommandEnvVariable.ExtensionSeqNumber])

                    status_contents = [{"status": {"status": ExtensionStatusValue.error, "code": fail_code,
                                                   "formattedMessage": {"message": fail_code, "lang": "en-US"}}}]
                    fileutil.write_file(os.path.join(env[ExtCommandEnvVariable.ExtensionPath], "status", status_file),
                                        json.dumps(status_contents))

                    invocation_records.append((full_name, handler_version, command_name))
                    # The return code is 0 but the status file should have the error, this it to test the scenario
                    # where the extensions return a success code but fail later.
                    return original_popen(['echo', "works"], *_, **kwargs)

            invocation_records.append((handler_name, handler_version, command_name))
            return original_popen(cmd, *_, **kwargs)

        with self._setup_test_env(mock_manifest=True) as (exthandlers_handler, protocol, no_of_extensions):
            with patch('azurelinuxagent.common.cgroupapi.subprocess.Popen', side_effect=mock_popen):
                exthandlers_handler.run()
                exthandlers_handler.report_ext_handlers_status()

                self.assertEqual(no_of_extensions,
                                 len(protocol.aggregate_status['aggregateStatus']['handlerAggregateStatus']),
                                 "incorrect extensions reported")

                # Since we're writing error status for firstExtension, only the firstExtension should be invoked and
                # everything else should be skipped
                expected_invocations = [
                    ('OSTCExtensions.ExampleHandlerLinux.firstExtension', '1.0.0', ExtensionCommandNames.INSTALL),
                    ('OSTCExtensions.ExampleHandlerLinux.firstExtension', '1.0.0', ExtensionCommandNames.ENABLE)]
                self.assertEqual(invocation_records, expected_invocations, "Invalid invocations found")

                err_msg = 'Dependent Extension OSTCExtensions.ExampleHandlerLinux.firstExtension did not succeed. Status was error'
                self.__assert_invalid_status_scenario(protocol, fail_code, mc_status="Ready",
                                                      mc_message="Plugin enabled", err_msg=err_msg)



#
# Copyright 2018 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Requires Python 2.6+ and Openssl 1.0+
#

import time

import azurelinuxagent.common.logger as logger
import azurelinuxagent.common.utils.fileutil as fileutil
import azurelinuxagent.common.utils.shellutil as shellutil  # pylint: disable=W0611
from azurelinuxagent.common.exception import OSUtilError  # pylint: disable=W0611
from azurelinuxagent.common.future import ustr  # pylint: disable=W0611
from azurelinuxagent.common.osutil.default import DefaultOSUtil


class SUSE11OSUtil(DefaultOSUtil):
    def __init__(self):
        super(SUSE11OSUtil, self).__init__()
        self.jit_enabled = True
        self.dhclient_name = 'dhcpcd'

    def set_hostname(self, hostname):
        fileutil.write_file('/etc/HOSTNAME', hostname)
        self._run_command_without_raising(["hostname", hostname], log_error=False)

    def get_dhcp_pid(self):
        return self._get_dhcp_pid(["pidof", self.dhclient_name])

    def is_dhcp_enabled(self):
        return True

    def stop_dhcp_service(self):
        self._run_command_without_raising(["/sbin/service", self.dhclient_name, "stop"], log_error=False)

    def start_dhcp_service(self):
        self._run_command_without_raising(["/sbin/service", self.dhclient_name, "start"], log_error=False)

    def start_network(self):
        self._run_command_without_raising(["/sbin/service", "network", "start"], log_error=False)

    def restart_ssh_service(self):
        self._run_command_without_raising(["/sbin/service", "sshd", "restart"], log_error=False)

    def stop_agent_service(self):
        self._run_command_without_raising(["/sbin/service", self.service_name, "stop"], log_error=False)

    def start_agent_service(self):
        self._run_command_without_raising(["/sbin/service", self.service_name, "start"], log_error=False)

    def register_agent_service(self):
        self._run_command_without_raising(["/sbin/insserv", self.service_name], log_error=False)

    def unregister_agent_service(self):
        self._run_command_without_raising(["/sbin/insserv", "-r", self.service_name], log_error=False)


class SUSEOSUtil(SUSE11OSUtil):
    def __init__(self):
        super(SUSEOSUtil, self).__init__()
        self.dhclient_name = 'wickedd-dhcp4'

    def publish_hostname(self, hostname):
        self.set_dhcp_hostname(hostname)
        self.set_hostname_record(hostname)
        ifname = self.get_if_name()
        # To push the hostname to the dhcp server we do not need to
        # bring down the interface, just make the make ifup do whatever is
        # necessary
        self.ifup(ifname)

    def ifup(self, ifname, retries=3, wait=5):
        logger.info('Interface {0} bounce with ifup'.format(ifname))
        retry_limit=retries+1
        for attempt in range(1, retry_limit):
            try:
                shellutil.run_command(['ifup', ifname], log_error=True)
            except Exception:
                if attempt < retry_limit:
                    logger.info("retrying in {0} seconds".format(wait))
                    time.sleep(wait)
                else:
                    logger.warn("exceeded restart retries")

    @staticmethod
    def get_systemd_unit_file_install_path():
        return "/usr/lib/systemd/system"

    def set_hostname(self, hostname):
        self._run_command_without_raising(
            ["hostnamectl", "set-hostname", hostname], log_error=False
        )

    def set_dhcp_hostname(self, hostname):
        dhcp_config_file_path = '/etc/sysconfig/network/dhcp'
        hostname_send_setting = fileutil.get_line_startingwith(
            'DHCLIENT_HOSTNAME_OPTION', dhcp_config_file_path
        )
        if hostname_send_setting:
            value = hostname_send_setting.split('=')[-1]
            # wicked's source accepts values with double quotes, single quotes, and no quotes at all.
            if value in ('"AUTO"', "'AUTO'", 'AUTO') or value == '"{0}"'.format(hostname):
                # Return if auto send host-name is configured or the current
                # hostname is already set up to be sent
                return
            else:
                # Do not use update_conf_file as it moves the setting to the
                # end of the file separating it from the contextual comment
                new_conf = []
                dhcp_conf = fileutil.read_file(
                    dhcp_config_file_path).split('\n')
                for entry in dhcp_conf:
                    if entry.startswith('DHCLIENT_HOSTNAME_OPTION'):
                        new_conf.append(
                           'DHCLIENT_HOSTNAME_OPTION="{0}"'. format(hostname)
                        )
                        continue
                    new_conf.append(entry)
                fileutil.write_file(dhcp_config_file_path, '\n'.join(new_conf))
        else:
            fileutil.append_file(
                dhcp_config_file_path,
                'DHCLIENT_HOSTNAME_OPTION="{0}"'. format(hostname)
            )

    def stop_dhcp_service(self):
        self._run_command_without_raising(["systemctl", "stop", "{}.service".format(self.dhclient_name)],
                                          log_error=False)

    def start_dhcp_service(self):
        self._run_command_without_raising(["systemctl", "start", "{}.service".format(self.dhclient_name)],
                                          log_error=False)

    def start_network(self):
        self._run_command_without_raising(["systemctl", "start", "network.service"], log_error=False)

    def restart_ssh_service(self):
        self._run_command_without_raising(["systemctl", "restart", "sshd.service"], log_error=False)

    def stop_agent_service(self):
        self._run_command_without_raising(["systemctl", "stop", "{}.service".format(self.service_name)],
                                          log_error=False)

    def start_agent_service(self):
        self._run_command_without_raising(["systemctl", "start", "{}.service".format(self.service_name)],
                                          log_error=False)

    def register_agent_service(self):
        self._run_command_without_raising(["systemctl", "enable", "{}.service".format(self.service_name)],
                                          log_error=False)

    def unregister_agent_service(self):
        self._run_command_without_raising(["systemctl", "disable", "{}.service".format(self.service_name)],
                                          log_error=False)



#!/usr/bin/env python3

# Microsoft Azure Linux Agent
#
# Copyright 2018 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from subprocess import Popen, PIPE
from typing import Any


class CommandError(Exception):
    """
    Exception raised by run_command when the command returns an error
    """
    def __init__(self, command: Any, exit_code: int, stdout: str, stderr: str):
        super().__init__(f"'{command}' failed (exit code: {exit_code}): {stderr}")
        self.command: Any = command
        self.exit_code: int = exit_code
        self.stdout: str = stdout
        self.stderr: str = stderr

    def __str__(self):
        return f"'{self.command}' failed (exit code: {self.exit_code})\nstdout:\n{self.stdout}\nstderr:\n{self.stderr}\n"


def run_command(command: Any, shell=False) -> str:
    """
    This function is a thin wrapper around Popen/communicate in the subprocess module. It executes the given command
    and returns its stdout. If the command returns a non-zero exit code, the function raises a RunCommandException.

    Similarly to Popen, the 'command' can be a string or a list of strings, and 'shell' indicates whether to execute
    the command through the shell.

    NOTE: The command's stdout and stderr are read as text streams.
    """
    process = Popen(command, stdout=PIPE, stderr=PIPE, shell=shell, text=True)

    stdout, stderr = process.communicate()

    if process.returncode != 0:
        raise CommandError(command, process.returncode, stdout, stderr)

    return stdout




# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
import logging
import re
import urllib.parse

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Type

# E0401: Unable to import 'dataclasses_json' (import-error)
from dataclasses_json import dataclass_json  # pylint: disable=E0401

# Disable those warnings, since 'lisa' is an external, non-standard, dependency
#     E0401: Unable to import 'lisa' (import-error)
#     etc
from lisa import schema  # pylint: disable=E0401
from lisa.combinator import Combinator  # pylint: disable=E0401
from lisa.util import field_metadata  # pylint: disable=E0401

from tests_e2e.orchestrator.lib.agent_test_loader import AgentTestLoader, VmImageInfo


@dataclass_json()
@dataclass
class AgentTestSuitesCombinatorSchema(schema.Combinator):
    test_suites: str = field(
        default_factory=str, metadata=field_metadata(required=True)
    )
    cloud: str = field(
        default_factory=str, metadata=field_metadata(required=True)
    )
    location: str = field(
        default_factory=str, metadata=field_metadata(required=True)
    )
    image: str = field(
        default_factory=str, metadata=field_metadata(required=False)
    )
    vm_size: str = field(
        default_factory=str, metadata=field_metadata(required=False)
    )
    vm_name: str = field(
        default_factory=str, metadata=field_metadata(required=False)
    )


class AgentTestSuitesCombinator(Combinator):
    """
    The "agent_test_suites" combinator returns a list of variables that specify the environments (i.e. test VMs) that the agent
    test suites must be executed on:

        * c_env_name: Unique name for the environment, e.g. "0001-com-ubuntu-server-focal-20_04-lts-westus2"
        * c_marketplace_image: e.g. "Canonical UbuntuServer 18.04-LTS latest",
        * c_location: e.g. "westus2",
        * c_vm_size: e.g. "Standard_D2pls_v5"
        * c_vhd: e.g "https://rhel.blob.core.windows.net/images/RHEL_8_Standard-8.3.202006170423.vhd?se=..."
        * c_test_suites: e.g. [AgentBvt, FastTrack]

    (c_marketplace_image, c_location, c_vm_size) and vhd are mutually exclusive and define the environment (i.e. the test VM)
    in which the test will be executed. c_test_suites defines the test suites that should be executed in that
    environment.

    The 'vm_name' runbook parameter can be used to execute the test suites on an existing VM. In that case, the combinator
    generates a single item with these variables:

        * c_env_name: Name for the environment, same as vm_name
        * c_vm_name:  Name of the test VM
        * c_location: Location of the test VM e.g. "westus2",
        * c_test_suites: e.g. [AgentBvt, FastTrack]
    """
    def __init__(self, runbook: AgentTestSuitesCombinatorSchema) -> None:
        super().__init__(runbook)
        if self.runbook.cloud not in self._DEFAULT_LOCATIONS:
            raise Exception(f"Invalid cloud: {self.runbook.cloud}")

        if self.runbook.vm_name != '' and (self.runbook.image != '' or self.runbook.vm_size != ''):
            raise Exception("Invalid runbook parameters: When 'vm_name' is specified, 'image' and 'vm_size' should not be specified.")

        if self.runbook.vm_name != '':
            self._environments = self.create_environment_for_existing_vm()
        else:
            self._environments = self.create_environment_list()
        self._index = 0


    @classmethod
    def type_name(cls) -> str:
        return "agent_test_suites"

    @classmethod
    def type_schema(cls) -> Type[schema.TypedSchema]:
        return AgentTestSuitesCombinatorSchema

    def _next(self) -> Optional[Dict[str, Any]]:
        result: Optional[Dict[str, Any]] = None
        if self._index < len(self._environments):
            result = self._environments[self._index]
            self._index += 1
        return result

    _DEFAULT_LOCATIONS = {
        "china": "china north 2",
        "government": "usgovarizona",
        "public": "westus2"
    }

    def create_environment_for_existing_vm(self) -> List[Dict[str, Any]]:
        loader = AgentTestLoader(self.runbook.test_suites)

        environment: List[Dict[str, Any]] = [
            {
                "c_env_name": self.runbook.vm_name,
                "c_vm_name": self.runbook.vm_name,
                "c_location": self.runbook.location,
                "c_test_suites": loader.test_suites,
            }
        ]

        log: logging.Logger = logging.getLogger("lisa")
        log.info("******** Environment for existing VMs *****")
        log.info(
            "{  c_env_name: '%s', c_vm_name: '%s', c_location: '%s', c_test_suites: '%s' }",
            environment[0]['c_env_name'], environment[0]['c_vm_name'], environment[0]['c_location'], [s.name for s in environment[0]['c_test_suites']])
        log.info("***************************")

        return environment

    def create_environment_list(self) -> List[Dict[str, Any]]:
        loader = AgentTestLoader(self.runbook.test_suites)

        #
        # If the runbook provides any of 'image', 'location', or 'vm_size', those values
        # override any configuration values on the test suite.
        #
        # Check 'images' first and add them to 'runbook_images', if any
        #
        if self.runbook.image == "":
            runbook_images = []
        else:
            runbook_images = loader.images.get(self.runbook.image)
            if runbook_images is None:
                if not self._is_urn(self.runbook.image) and not self._is_vhd(self.runbook.image):
                    raise Exception(f"The 'image' parameter must be an image or image set name, a urn, or a vhd: {self.runbook.image}")
                i = VmImageInfo()
                i.urn = self.runbook.image  # Note that this could be a URN or the URI for a VHD
                i.locations = []
                i.vm_sizes = []
                runbook_images = [i]

        #
        # Now walk through all the test_suites and create a list of the environments (test VMs) that need to be created.
        #
        environment_list: List[Dict[str, Any]] = []
        shared_environments: Dict[str, Dict[str, Any]] = {}

        for suite_info in loader.test_suites:
            if len(runbook_images) > 0:
                images_info = runbook_images
            else:
                # The test suite may be referencing multiple image sets, and sets can intersect, so we need to ensure
                # we eliminate any duplicates.
                unique_images: Dict[str, str] = {}
                for image in suite_info.images:
                    for i in loader.images[image]:
                        unique_images[i] = i
                images_info = unique_images.values()

            for image in images_info:
                # The URN can actually point to a VHD if the runbook provided a VHD in the 'images' parameter
                if self._is_vhd(image.urn):
                    marketplace_image = ""
                    vhd = image.urn
                    name = "vhd"
                else:
                    marketplace_image = image.urn
                    vhd = ""
                    match = AgentTestSuitesCombinator._URN.match(image.urn)
                    if match is None:
                        raise Exception(f"Invalid URN: {image.urn}")
                    name = f"{match.group('offer')}-{match.group('sku')}"

                # If the runbook specified a location, use it. Then try the suite location, if any. Otherwise, check if the image specifies
                # a list of locations and use any of them. If no location is specified so far, use the default.
                if self.runbook.location != "":
                    location = self.runbook.location
                elif suite_info.location != '':
                    location = suite_info.location
                elif len(image.locations) > 0:
                    location = image.locations[0]
                else:
                    location = AgentTestSuitesCombinator._DEFAULT_LOCATIONS[self.runbook.cloud]

                # If the runbook specified a VM size, use it. Else if the image specifies a list of VM sizes, use any of them. Otherwise,
                # set the size to empty and let LISA choose it.
                if self.runbook.vm_size != '':
                    vm_size = self.runbook.vm_size
                elif len(image.vm_sizes) > 0:
                    vm_size = image.vm_sizes[0]
                else:
                    vm_size = ""

                if suite_info.owns_vm:
                    # create an environment for exclusive use by this suite
                    environment_list.append({
                        "c_marketplace_image": marketplace_image,
                        "c_location": location,
                        "c_vm_size": vm_size,
                        "c_vhd": vhd,
                        "c_test_suites": [suite_info],
                        "c_env_name": f"{name}-{suite_info.name}"
                    })
                else:
                    # add this suite to the shared environments
                    key: str = f"{name}-{location}"
                    if key in shared_environments:
                        shared_environments[key]["c_test_suites"].append(suite_info)
                    else:
                        shared_environments[key] = {
                            "c_marketplace_image": marketplace_image,
                            "c_location": location,
                            "c_vm_size": vm_size,
                            "c_vhd": vhd,
                            "c_test_suites": [suite_info],
                            "c_env_name": key
                        }

        environment_list.extend(shared_environments.values())

        log: logging.Logger = logging.getLogger("lisa")
        log.info("******** Environments *****")
        for e in environment_list:
            log.info(
                "{ c_marketplace_image: '%s', c_location: '%s', c_vm_size: '%s', c_vhd: '%s', c_test_suites: '%s', c_env_name: '%s' }",
                e['c_marketplace_image'], e['c_location'], e['c_vm_size'], e['c_vhd'], [s.name for s in e['c_test_suites']], e['c_env_name'])
        log.info("***************************")

        return environment_list

    _URN = re.compile(r"(?P<publisher>[^\s:]+)[\s:](?P<offer>[^\s:]+)[\s:](?P<sku>[^\s:]+)[\s:](?P<version>[^\s:]+)")

    @staticmethod
    def _is_urn(urn: str) -> bool:
        # URNs can be given as '<Publisher> <Offer> <Sku> <Version>' or '<Publisher>:<Offer>:<Sku>:<Version>'
        return AgentTestSuitesCombinator._URN.match(urn) is not None

    @staticmethod
    def _is_vhd(vhd: str) -> bool:
        # VHDs are given as URIs to storage; do some basic validation, not intending to be exhaustive.
        parsed = urllib.parse.urlparse(vhd)
        return parsed.scheme == 'https' and parsed.netloc != "" and parsed.path != ""
